from unittest.mock import call, patch

import pytest
import zigpy.types as t

from universal_silabs_flasher.cpc import (
    CPCProtocol,
    CPCTransportFrame,
    PropertyCommand,
    UnnumberedFrame,
)
from universal_silabs_flasher.cpc_types import (
    EndpointId,
    PropertyId,
    UnnumberedFrameCommandId,
)

FRAME1 = CPCTransportFrame(
    endpoint=EndpointId.SYSTEM,
    control=t.uint8_t(196),
    payload=UnnumberedFrame(
        command_id=UnnumberedFrameCommandId.PROP_VALUE_IS,
        command_seq=t.uint8_t(0),
        payload=PropertyCommand(
            property_id=PropertyId.SECONDARY_CPC_VERSION,
            value=b"\x04\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00",
        ),
    ),
)
FRAME2 = CPCTransportFrame(
    endpoint=EndpointId.SYSTEM,
    control=t.uint8_t(196),
    payload=UnnumberedFrame(
        command_id=UnnumberedFrameCommandId.PROP_VALUE_IS,
        command_seq=t.uint8_t(1),
        payload=PropertyCommand(
            property_id=PropertyId.SECONDARY_APP_VERSION, value=b"4.4.3-0368642c\x00"
        ),
    ),
)


def test_cpc_serialization() -> None:
    assert FRAME1.serialize() == (
        b"\x14\x00\x16\x00\xc4W\xe5\x06\x00\x10\x00\x03\x00\x00\x00\x04\x00\x00\x00\x04"
        b"\x00\x00\x00\x03\x00\x00\x00}>"
    )
    assert FRAME2.serialize() == (
        b"\x14\x00\x19\x00\xc4f\xc9\x06\x01\x13\x00\x04\x00\x00\x004.4.3-0368642c\x00"
        b"\x0b\xba"
    )


@pytest.mark.parametrize(
    "chunks",
    [
        # One at a time
        [FRAME1.serialize(), FRAME2.serialize()],
        # Byte at a time
        [bytes([b]) for b in FRAME1.serialize() + FRAME2.serialize()],
        # Concatenated
        [FRAME1.serialize() + FRAME2.serialize()],
    ],
)
def test_cpc_deserialization(chunks: list[bytes]) -> None:
    cpc = CPCProtocol()

    with patch.object(cpc, "frame_received") as mock_frame_received:
        for chunk in chunks:
            cpc.data_received(chunk)

    assert mock_frame_received.mock_calls == [
        call(FRAME1),
        call(FRAME2),
    ]


def test_cpc_bad_buffer_deserialization() -> None:
    cpc = CPCProtocol()

    with patch.object(cpc, "frame_received") as mock_frame_received:
        cpc.data_received(b"aaaaaaaaaaaaaaaaaaaaa\r\n")

    assert mock_frame_received.mock_calls == []
    assert cpc._buffer == b""



from __future__ import annotations

from collections.abc import Sequence
import dataclasses
import enum


class FirmwareImageType(enum.Enum):
    ZIGBEE_NCP = "zigbee_ncp"
    ZIGBEE_ROUTER = "zigbee_router"
    OPENTHREAD_RCP = "openthread_rcp"
    ZWAVE_NCP = "zwave_ncp"
    BOOTLOADER = "bootloader"
    MULTIPAN = "multipan"

    UNKNOWN = "unknown"


LEGACY_FIRMWARE_TYPE_REMAPPING = {
    "ncp-uart-hw": FirmwareImageType.ZIGBEE_NCP,
    "ncp-uart-sw": FirmwareImageType.ZIGBEE_NCP,
    "rcp-uart-802154": FirmwareImageType.MULTIPAN,
    "ot-rcp": FirmwareImageType.OPENTHREAD_RCP,
    "z-wave": FirmwareImageType.ZWAVE_NCP,
    "gecko-bootloader": FirmwareImageType.BOOTLOADER,
}


class ApplicationType(enum.Enum):
    GECKO_BOOTLOADER = "bootloader"
    CPC = "cpc"
    EZSP = "ezsp"
    SPINEL = "spinel"
    ROUTER = "router"
    ZWAVE = "zwave"


FW_IMAGE_TYPE_TO_APPLICATION_TYPE = {
    FirmwareImageType.ZIGBEE_NCP: ApplicationType.EZSP,
    FirmwareImageType.MULTIPAN: ApplicationType.CPC,
    FirmwareImageType.OPENTHREAD_RCP: ApplicationType.SPINEL,
    FirmwareImageType.BOOTLOADER: ApplicationType.GECKO_BOOTLOADER,
    FirmwareImageType.ZIGBEE_ROUTER: ApplicationType.ROUTER,
    FirmwareImageType.ZWAVE_NCP: ApplicationType.ZWAVE,
}


DEFAULT_PROBE_METHODS = (
    (ApplicationType.GECKO_BOOTLOADER, 115200),
    (ApplicationType.EZSP, 115200),
    (ApplicationType.EZSP, 460800),
    (ApplicationType.SPINEL, 460800),
    (ApplicationType.CPC, 460800),
    (ApplicationType.CPC, 115200),
    (ApplicationType.CPC, 230400),
    (ApplicationType.ROUTER, 115200),
)


class ResetTarget(enum.Enum):
    YELLOW = "yellow"
    IHOST = "ihost"
    SLZB07 = "slzb07"
    RTS_DTR = "rts_dtr"
    BAUDRATE = "baudrate"


@dataclasses.dataclass
class GpioPattern:
    pins: dict[int, bool]
    delay_after: float


@dataclasses.dataclass
class GpioResetConfig:
    chip: str | None
    chip_type: str | None
    pattern: list[GpioPattern]


@dataclasses.dataclass
class ModemPinPattern:
    pins: dict[str, bool]
    delay_after: float


@dataclasses.dataclass
class ModemPinResetConfig:
    pattern: list[ModemPinPattern]


@dataclasses.dataclass
class BaudrateResetConfig:
    baudrates: Sequence[int]
    delay_after_each: float
    delay_after_final: float
    command: bytes | None = None


# fmt: off
RESET_CONFIGS: dict[
    ResetTarget, GpioResetConfig | ModemPinResetConfig | BaudrateResetConfig
] = {
    ResetTarget.YELLOW: GpioResetConfig(
        chip="/dev/gpiochip0",
        chip_type=None,
        pattern=[
            GpioPattern(pins={24: True,  25: True},  delay_after=0.1),
            GpioPattern(pins={24: False, 25: False}, delay_after=0.1),
            GpioPattern(pins={24: False, 25: True},  delay_after=0.1),
            GpioPattern(pins={24: True,  25: True},  delay_after=0.0),
        ],
    ),
    ResetTarget.IHOST: GpioResetConfig(
        chip="/dev/gpiochip1",
        chip_type=None,
        pattern=[
            GpioPattern(pins={26: True,  27: True},  delay_after=0.1),
            GpioPattern(pins={26: False, 27: False}, delay_after=0.1),
            GpioPattern(pins={26: True,  27: False}, delay_after=0.1),
            GpioPattern(pins={26: True,  27: True},  delay_after=0.0),
        ]
    ),
    ResetTarget.SLZB07: GpioResetConfig(
        chip=None,
        chip_type="cp210x",
        pattern=[
            GpioPattern(pins={4: True,  5: True},  delay_after=0.1),
            GpioPattern(pins={4: False, 5: False}, delay_after=0.1),
            GpioPattern(pins={4: True,  5: False}, delay_after=0.1),
            GpioPattern(pins={4: True,  5: True},  delay_after=0.0),
        ]
    ),
    ResetTarget.RTS_DTR: ModemPinResetConfig(
        pattern=[
            ModemPinPattern(pins={"dtr": False, "rts": True},  delay_after=0.1),
            ModemPinPattern(pins={"dtr": True,  "rts": False}, delay_after=0.5),
            ModemPinPattern(pins={"dtr": False, "rts": False}, delay_after=0.0),
        ]
    ),
    ResetTarget.BAUDRATE: BaudrateResetConfig(
        baudrates=[150, 300, 1200],
        delay_after_each=0.1,
        delay_after_final=0.5,
        command=b"BZ",
    ),
}
# fmt: on



from __future__ import annotations

import asyncio
import logging
import pathlib
import sys
from unittest.mock import MagicMock, call

import pytest

from universal_silabs_flasher.common import crc16_ccitt, pad_to_multiple
from universal_silabs_flasher.gecko_bootloader import (
    XMODEM_BLOCK_SIZE,
    GeckoBootloaderOption,
    GeckoBootloaderProtocol,
    ReceiverCancelled,
    UploadError,
    XModemPacketType,
)

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout  # pragma: no cover
else:
    from asyncio import timeout as asyncio_timeout  # pragma: no cover


from .common import PairedTransport

_LOGGER = logging.getLogger(__name__)

FULL_FIRMWARE = pad_to_multiple(
    (
        pathlib.Path(__file__).parent / "firmwares/skyconnect_zigbee_ncp_7.4.4.0.gbl"
    ).read_bytes(),
    XMODEM_BLOCK_SIZE,
    b"\xff",
)

FIRMWARE = FULL_FIRMWARE[: 100 * XMODEM_BLOCK_SIZE]

UPLOAD_COMPLETION_TIMEOUT = 2.5


class Conversation(asyncio.Protocol):
    """A helper to script a conversation with a protocol."""

    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
        self.loop = loop
        self.transport: PairedTransport | None = None
        self._reader = asyncio.StreamReader(loop=loop)

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        assert isinstance(transport, PairedTransport)
        self.transport = transport

    def data_received(self, data: bytes) -> None:
        _LOGGER.debug("Conversation received: %r", data)
        self._reader.feed_data(data)

    def connection_lost(self, exc: Exception | None) -> None:
        self._reader.feed_eof()

    async def send(self, data: bytes) -> None:
        assert self.transport is not None
        self.transport.write(data)

    async def send_menu(self) -> None:
        menu = (
            "\r\nGecko Bootloader v1.12.1\r\n"
            "1. upload gbl\r\n"
            "2. run\r\n"
            "3. ebl info\r\n"
            "BL > "
        ).encode("ascii")
        await self.send(menu)

    async def send_ack(self) -> None:
        await self.send(bytes([XModemPacketType.ACK]))

    async def send_nak(self) -> None:
        await self.send(bytes([XModemPacketType.NAK]))

    async def send_can(self) -> None:
        await self.send(bytes([XModemPacketType.CAN]))

    async def send_upload_complete(self) -> None:
        await self.send(b"\r\nSerial upload complete\r\n")

    async def expect_command(
        self,
        command: GeckoBootloaderOption,
        timeout: float = 1.5,
    ) -> None:
        """Expect a command, preceded by a newline."""
        async with asyncio_timeout(timeout):
            data = await self._reader.read(len(command) + 1)
        assert data.endswith(command)

    async def expect_packet(self, number: int, timeout: float = 2.0) -> bytes:
        """Read and validate a full XMODEM packet, returning its payload."""
        # 3 bytes header, 128 bytes payload, 2 bytes CRC
        async with asyncio_timeout(timeout):
            data = await self._reader.read(133)

        assert data[0] == XModemPacketType.SOH
        assert data[1] == number
        assert data[2] == 0xFF - number

        payload = data[3:-2]
        crc = int.from_bytes(data[-2:], "big")
        assert crc16_ccitt(payload) == crc

        return payload

    async def expect_eot(self, timeout: float = 1.0) -> None:
        async with asyncio_timeout(timeout):
            data = await self._reader.read(1)
        assert data == bytes([XModemPacketType.EOT])


async def create_test_pair(
    chunk_size: int | None = None,
    aggregate_write_timeout: float | None = None,
) -> tuple[GeckoBootloaderProtocol, Conversation]:
    """Creates a connected pair of a client protocol and a conversation helper."""
    loop = asyncio.get_running_loop()
    client = GeckoBootloaderProtocol()
    conversation = Conversation(loop)

    client_transport = PairedTransport(
        conversation,
        loop,
        chunk_size=chunk_size,
        aggregate_write_timeout=aggregate_write_timeout,
    )
    server_transport = PairedTransport(
        client,
        loop,
        chunk_size=chunk_size,
        aggregate_write_timeout=aggregate_write_timeout,
    )

    client.connection_made(client_transport)
    conversation.connection_made(server_transport)

    return client, conversation


async def test_xmodem_happy_path() -> None:
    """Test a successful XMODEM transfer."""
    client, conversation = await create_test_pair()
    received_firmware = bytearray()
    progress_callback = MagicMock()

    # Start the upload and script the conversation
    upload_task = asyncio.create_task(
        client.upload_firmware(FULL_FIRMWARE, progress_callback=progress_callback)
    )

    # The client automatically queries for info, so reply with a menu
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    for i in range(len(FULL_FIRMWARE) // XMODEM_BLOCK_SIZE):
        payload = await conversation.expect_packet(number=(i + 1) & 0xFF)
        received_firmware.extend(payload)
        await conversation.send_ack()

    await conversation.expect_eot()
    await conversation.send_ack()
    await conversation.send_upload_complete()

    # Final menu prompt
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    async with asyncio_timeout(UPLOAD_COMPLETION_TIMEOUT):
        await upload_task

    assert received_firmware == FULL_FIRMWARE
    assert progress_callback.mock_calls == [
        call(i * XMODEM_BLOCK_SIZE, len(FULL_FIRMWARE))
        for i in range(1 + len(FULL_FIRMWARE) // XMODEM_BLOCK_SIZE)
    ]


async def test_xmodem_with_retries() -> None:
    """Test an XMODEM transfer with some retries."""
    client, conversation = await create_test_pair()
    received_firmware = bytearray()

    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE))

    # Initial info query
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    # Upload command
    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    # First packet is OK
    payload = await conversation.expect_packet(number=1)
    received_firmware.extend(payload)
    await conversation.send_ack()

    # Second packet is NAK'd once
    await conversation.expect_packet(number=2)
    await conversation.send_nak()
    payload = await conversation.expect_packet(number=2)
    received_firmware.extend(payload)
    await conversation.send_ack()

    # The rest are OK
    for i in range(2, len(FIRMWARE) // XMODEM_BLOCK_SIZE):
        payload = await conversation.expect_packet(number=(i + 1) & 0xFF)
        received_firmware.extend(payload)
        await conversation.send_ack()

    await conversation.expect_eot()
    await conversation.send_ack()
    await conversation.send_upload_complete()

    # Final menu prompt
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    async with asyncio_timeout(UPLOAD_COMPLETION_TIMEOUT):
        await upload_task

    assert received_firmware == FIRMWARE


async def test_xmodem_timeout() -> None:
    """Test an XMODEM transfer with a receive timeout on the server side."""
    client, conversation = await create_test_pair()
    received_firmware = bytearray()

    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE))

    # Initial info query
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    # Upload command
    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    # First packet is OK
    payload = await conversation.expect_packet(number=1)
    received_firmware.extend(payload)
    await conversation.send_ack()

    # Do not reply to the second packet to trigger a timeout
    await conversation.expect_packet(number=2)
    payload = await conversation.expect_packet(number=2, timeout=3.0)
    received_firmware.extend(payload)
    await conversation.send_ack()

    # The rest are OK
    for i in range(2, len(FIRMWARE) // XMODEM_BLOCK_SIZE):
        payload = await conversation.expect_packet(number=(i + 1) & 0xFF)
        received_firmware.extend(payload)
        await conversation.send_ack()

    await conversation.expect_eot()
    await conversation.send_ack()
    await conversation.send_upload_complete()

    # Final menu prompt
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    async with asyncio_timeout(UPLOAD_COMPLETION_TIMEOUT):
        await upload_task

    assert received_firmware == FIRMWARE


async def test_xmodem_cancellation() -> None:
    """Test an XMODEM transfer that is cancelled by the receiver."""
    client, conversation = await create_test_pair()

    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE))

    # Initial info query
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    # Upload command
    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    # First packet is cancelled
    await conversation.expect_packet(number=1)
    await conversation.send_can()

    with pytest.raises(ReceiverCancelled):
        async with asyncio_timeout(1):
            await upload_task


async def test_xmodem_too_many_retries() -> None:
    """Test an XMODEM transfer that fails after too many retries."""
    client, conversation = await create_test_pair()

    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE, max_failures=3))

    # Initial info query
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    # Upload command
    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    # Continuously NAK the first packet
    for i in range(4):
        await conversation.expect_packet(number=1)
        await conversation.send_nak()

    with pytest.raises(UploadError):
        await upload_task


async def test_xmodem_ack_with_garbage() -> None:
    """Test that the client handles invalid XMODEM responses."""
    client, conversation = await create_test_pair()

    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE))

    # Initial info query
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    # Upload command
    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    # First packet is OK
    await conversation.expect_packet(number=1)
    await conversation.send_ack()

    # Send invalid responses (not ACK, NAK, or CAN) - should retry and eventually fail
    for i in range(4):
        await conversation.expect_packet(number=2)
        await conversation.send(b"\xff")  # Invalid response

    # The upload fails due to too many retries
    with pytest.raises(UploadError, match="Received 3 consecutive failures"):
        await upload_task


async def test_xmodem_multiple_c_bytes() -> None:
    """Test that the client handles multiple `C` bytes from the bootloader."""
    client, conversation = await create_test_pair()
    received_firmware = bytearray()

    # Start the upload and script the conversation
    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE))

    # The client automatically queries for info, so reply with a menu
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    # Send multiple `C` bytes
    await conversation.send(b"CCC")

    for i in range(len(FIRMWARE) // XMODEM_BLOCK_SIZE):
        payload = await conversation.expect_packet(number=(i + 1) & 0xFF)
        received_firmware.extend(payload)
        await conversation.send_ack()

    await conversation.expect_eot()
    await conversation.send_ack()
    await conversation.send_upload_complete()

    # Final menu prompt
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    async with asyncio_timeout(UPLOAD_COMPLETION_TIMEOUT):
        await upload_task

    assert received_firmware == FIRMWARE


async def test_xmodem_spurious_ack() -> None:
    """Test that the client ignores a spurious ACK."""
    client, conversation = await create_test_pair()
    received_firmware = bytearray()

    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE))

    # Initial info query
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    # Upload command
    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    # Send a spurious ACK, client should ignore it and send packet 1
    await conversation.send_ack()

    # The rest of the transfer should proceed normally
    for i in range(len(FIRMWARE) // XMODEM_BLOCK_SIZE):
        payload = await conversation.expect_packet(number=(i + 1) & 0xFF)
        received_firmware.extend(payload)
        await conversation.send_ack()

    await conversation.expect_eot()
    await conversation.send_ack()
    await conversation.send_upload_complete()

    # Final menu prompt
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    async with asyncio_timeout(UPLOAD_COMPLETION_TIMEOUT):
        await upload_task

    assert received_firmware == FIRMWARE


async def test_xmodem_spurious_nak() -> None:
    """Test that the client ignores a spurious NAK."""
    client, conversation = await create_test_pair()
    received_firmware = bytearray()

    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE))

    # Initial info query
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    # Upload command
    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    # Send a spurious NAK, client should ignore it and send packet 1
    await conversation.send_nak()

    # The rest of the transfer should proceed normally
    for i in range(len(FIRMWARE) // XMODEM_BLOCK_SIZE):
        payload = await conversation.expect_packet(number=(i + 1) & 0xFF)
        received_firmware.extend(payload)
        await conversation.send_ack()

    await conversation.expect_eot()
    await conversation.send_ack()
    await conversation.send_upload_complete()

    # Final menu prompt
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    async with asyncio_timeout(UPLOAD_COMPLETION_TIMEOUT):
        await upload_task

    assert received_firmware == FIRMWARE


async def test_xmodem_task_cancellation() -> None:
    """Test that the client handles task cancellation."""
    client, conversation = await create_test_pair()

    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE))

    # Initial info query
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    # Upload command
    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    # First packet is OK
    await conversation.expect_packet(number=1)
    await conversation.send_ack()

    # Cancel the task during the transfer
    upload_task.cancel()

    with pytest.raises(asyncio.CancelledError):
        await upload_task


async def test_xmodem_connection_lost() -> None:
    """Test that the client handles connection loss during XMODEM transfer."""
    client, conversation = await create_test_pair()

    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE))

    # Initial info query
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    # Upload command
    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    # First packet is OK
    await conversation.expect_packet(number=1)
    await conversation.send_ack()

    # Second packet is sent, but connection is lost before response
    await conversation.expect_packet(number=2)

    # Simulate connection loss (disconnect) by calling connection_lost directly
    client.connection_lost(None)

    # The upload should fail with connection lost error
    with pytest.raises(RuntimeError, match="Connection has been lost"):
        await upload_task


async def test_xmodem_can_during_transfer() -> None:
    """Test that the client handles CAN (cancel) byte during transfer."""
    client, conversation = await create_test_pair()

    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE))

    # Initial info query
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    # Upload command
    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    # First packet is OK
    await conversation.expect_packet(number=1)
    await conversation.send_ack()

    # Second packet gets a CAN response
    await conversation.expect_packet(number=2)
    await conversation.send(bytes([XModemPacketType.CAN]))

    # The upload should fail with ReceiverCancelled
    with pytest.raises(ReceiverCancelled):
        await upload_task


async def test_xmodem_empty_buffer_during_transfer() -> None:
    """Test that empty data during XMODEM transfer is handled gracefully."""
    client, conversation = await create_test_pair()

    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE))

    # Initial info query
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    # Upload command
    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    # First packet is OK
    await conversation.expect_packet(number=1)
    # Send empty data (should be ignored due to empty buffer check)
    await conversation.send(b"")
    # Then send ACK
    await conversation.send_ack()

    # Continue with rest of transfer
    for i in range(1, len(FIRMWARE) // XMODEM_BLOCK_SIZE):
        await conversation.expect_packet(number=(i + 1) & 0xFF)
        await conversation.send_ack()

    await conversation.expect_eot()
    await conversation.send_ack()
    await conversation.send_upload_complete()
    await conversation.send_menu()

    async with asyncio_timeout(1):
        await upload_task


async def test_xmodem_abort_with_pending_timeout() -> None:
    """Test that abort properly cancels pending timeout."""
    client, conversation = await create_test_pair()

    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE))

    # Initial info query
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    # Upload command
    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    # First packet - send it but don't respond yet
    await conversation.expect_packet(number=1)

    # At this point, a timeout is pending. Manually trigger abort.
    # This exercises the timeout cancellation in _xmodem_abort()
    error = UploadError("Manual abort test")
    client._xmodem_abort(error)

    # The upload should fail with our error
    with pytest.raises(UploadError, match="Manual abort test"):
        await upload_task


async def test_xmodem_reverts_to_line_parsing() -> None:
    """Test that the client reverts to line-based parsing after an upload."""
    client, conversation = await create_test_pair()
    received_firmware = bytearray()

    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE))

    # Initial info query
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    # Upload command
    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    # Full transfer
    for i in range(len(FIRMWARE) // XMODEM_BLOCK_SIZE):
        payload = await conversation.expect_packet(number=(i + 1) & 0xFF)
        received_firmware.extend(payload)
        await conversation.send_ack()

    await conversation.expect_eot()
    await conversation.send_ack()

    # Send the upload complete and menu messages back-to-back
    await conversation.send_upload_complete()
    await conversation.send_menu()

    # The upload task should complete successfully
    async with asyncio_timeout(1):
        await upload_task

    assert received_firmware == FIRMWARE
    assert client._state_machine.state == "in_menu"


async def test_xmodem_reverts_to_line_parsing_byte_by_byte() -> None:
    """Test line-based parsing after upload with byte-by-byte delivery."""
    client, conversation = await create_test_pair(chunk_size=1)
    received_firmware = bytearray()

    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE))

    # Initial info query
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    # Upload command
    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    # Full transfer
    for i in range(len(FIRMWARE) // XMODEM_BLOCK_SIZE):
        payload = await conversation.expect_packet(number=(i + 1) & 0xFF)
        received_firmware.extend(payload)
        await conversation.send_ack()

    await conversation.expect_eot()
    await conversation.send_ack()

    # Send the upload complete and menu messages back-to-back
    await conversation.send_upload_complete()
    await conversation.send_menu()

    # The upload task should complete successfully
    async with asyncio_timeout(1):
        await upload_task

    assert received_firmware == FIRMWARE
    assert client._state_machine.state == "in_menu"


async def test_xmodem_reverts_to_line_parsing_with_write_aggregation() -> None:
    """Test line-based parsing after upload with write aggregation."""
    client, conversation = await create_test_pair(aggregate_write_timeout=0.05)
    received_firmware = bytearray()

    upload_task = asyncio.create_task(client.upload_firmware(FIRMWARE))

    # Initial info query
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()

    # Upload command
    await conversation.expect_command(GeckoBootloaderOption.UPLOAD_FIRMWARE)
    await conversation.send(b"C")

    # Full transfer
    for i in range(len(FIRMWARE) // XMODEM_BLOCK_SIZE):
        payload = await conversation.expect_packet(number=(i + 1) & 0xFF)
        received_firmware.extend(payload)
        await conversation.send_ack()

    await conversation.expect_eot()
    await conversation.send_ack()

    # Send the upload complete and menu messages back-to-back
    # With write aggregation, these should be buffered and delivered together
    await conversation.send_upload_complete()
    await conversation.send_menu()

    # The upload task should complete successfully
    async with asyncio_timeout(1):
        await upload_task

    assert received_firmware == FIRMWARE
    assert client._state_machine.state == "in_menu"


async def test_parser_needs_more_data() -> None:
    """Test that parser correctly waits when it needs more data."""
    client, conversation = await create_test_pair()

    # Start the ebl_info query
    info_task = asyncio.create_task(client.ebl_info())

    # Expect the command
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)

    # Send partial menu data that won't match the regex yet
    partial_menu = b"\r\nGecko Bootloader v1.12.1\r\n1. upload"
    await conversation.send(partial_menu)

    # Give the parser a moment to process
    await asyncio.sleep(0.1)

    # The state should still be WAITING_FOR_MENU since we don't have a complete menu
    assert client._state_machine.state == "waiting_for_menu"

    # Now send the rest of the menu
    rest_of_menu = b" gbl\r\n2. run\r\n3. ebl info\r\nBL > "
    await conversation.send(rest_of_menu)

    # The info task should complete
    async with asyncio_timeout(1):
        await info_task


async def test_spurious_data_in_menu() -> None:
    """Test that spurious data while in IN_MENU state is ignored."""
    client, conversation = await create_test_pair()

    # Get into IN_MENU state
    info_task = asyncio.create_task(client.ebl_info())
    await conversation.expect_command(GeckoBootloaderOption.EBL_INFO)
    await conversation.send_menu()
    await info_task

    # Now we're in IN_MENU state. Send some spurious data.
    await conversation.send(b"xyz")

    # Give the parser a moment to process
    await asyncio.sleep(0.1)

    # Should still be in IN_MENU, just ignoring the spurious data
    assert client._state_machine.state == "in_menu"



from __future__ import annotations

import asyncio
import dataclasses
import logging
import typing

from typing_extensions import Self
from zigpy.serial import SerialProtocol
import zigpy.types as t

from .common import BufferTooShort, Version, asyncio_timeout

_LOGGER = logging.getLogger(__name__)

COMMAND_TIMEOUT = 2


class MessageType(t.enum8):
    REQUEST = 0x00
    RESPONSE = 0x01


class FunctionID(t.enum8):
    SERIAL_API_GET_CAPABILITIES = 0x07
    SERIAL_API_ENTER_BOOTLOADER = 0x27


SOF = 0x01
ACK = 0x06
NAK = 0x15
CAN = 0x18


@dataclasses.dataclass(frozen=True)
class ZWaveFrame:
    type: MessageType
    function_id: FunctionID
    data: bytes

    @classmethod
    def deserialize(cls, data: bytes | bytearray) -> tuple[Self, bytes | bytearray]:
        if len(data) < 2:
            raise BufferTooShort()

        if data[0] != SOF:
            raise ValueError(f"Expected SOF, got 0x{data[0]:02x}")

        length = data[1]

        if len(data) < 2 + length:
            raise BufferTooShort()

        expected_checksum = data[1 + length]
        computed_checksum = 0xFF

        for b in data[1 : 1 + length]:
            computed_checksum ^= b

        if computed_checksum != expected_checksum:
            raise ValueError(
                f"Checksum mismatch: expected 0x{computed_checksum:02x},"
                f" got 0x{expected_checksum:02x}"
            )

        msg_type = MessageType(data[2])  # type: ignore[no-untyped-call]
        function_id = FunctionID(data[3])  # type: ignore[no-untyped-call]
        payload = bytes(data[4 : 1 + length])

        return (
            cls(type=msg_type, function_id=function_id, data=payload),
            data[2 + length :],
        )

    def serialize(self) -> bytes:
        length = 1 + 1 + len(self.data) + 1
        data = (
            bytes([length])
            + self.type.serialize()
            + self.function_id.serialize()
            + self.data
        )

        checksum = 0xFF
        for b in data:
            checksum ^= b

        return bytes([SOF]) + data + bytes([checksum])


class ZWaveProtocol(SerialProtocol):
    _buffer: bytearray

    def __init__(self) -> None:
        super().__init__()
        self._pending_frames: dict[FunctionID, asyncio.Future[ZWaveFrame]] = {}

    def send_data(self, data: bytes) -> None:
        assert self._transport is not None
        _LOGGER.debug("Sending data %s", data)
        self._transport.write(data)

    def data_received(self, data: bytes) -> None:
        super().data_received(data)

        while self._buffer:
            byte = self._buffer[0]

            if byte == ACK:
                _LOGGER.debug("Received ACK")
                self._buffer = self._buffer[1:]
                continue

            if byte in (NAK, CAN):
                _LOGGER.debug("Received 0x%02x", byte)
                self._buffer = self._buffer[1:]
                continue

            if byte != SOF:
                _LOGGER.debug("Discarding unexpected byte 0x%02x", byte)
                self._buffer = self._buffer[1:]
                continue

            try:
                frame, new_buffer = ZWaveFrame.deserialize(self._buffer)
            except BufferTooShort:
                break
            except ValueError as e:
                _LOGGER.debug("Failed to parse buffer %r: %r", self._buffer, e)
                self._buffer = self._buffer[1:]
            else:
                self._buffer = typing.cast(bytearray, new_buffer)
                self.send_data(bytes([ACK]))
                self.frame_received(frame)

    def frame_received(self, frame: ZWaveFrame) -> None:
        _LOGGER.debug("Received frame %r", frame)

        if frame.type != MessageType.RESPONSE:
            return

        if frame.function_id not in self._pending_frames:
            _LOGGER.debug("Received unsolicited response for %r", frame.function_id)
            return

        future = self._pending_frames[frame.function_id]

        if future.done():
            _LOGGER.debug("Ignoring duplicate response for %r", frame.function_id)
            return

        future.set_result(frame)

    async def send_command(
        self,
        function_id: FunctionID,
        data: bytes = b"",
        *,
        retries: int = 2,
        timeout: float = COMMAND_TIMEOUT,
        retry_delay: float = 0.1,
    ) -> ZWaveFrame:
        assert function_id not in self._pending_frames

        future: asyncio.Future[ZWaveFrame] = asyncio.get_running_loop().create_future()
        self._pending_frames[function_id] = future

        try:
            for attempt in range(retries + 1):
                frame = ZWaveFrame(
                    type=MessageType.REQUEST, function_id=function_id, data=data
                )

                _LOGGER.debug("Sending frame %r", frame)
                self.send_data(frame.serialize())

                try:
                    async with asyncio_timeout(timeout):
                        return await future
                except asyncio.TimeoutError:
                    _LOGGER.debug(
                        "Failed to send %r, trying again in %0.2fs (attempt %s of %s)",
                        frame,
                        retry_delay,
                        attempt + 1,
                        retries + 1,
                    )

                    if attempt >= retries:
                        raise

                    await asyncio.sleep(retry_delay)
        finally:
            self._pending_frames.pop(function_id, None)

        # Unreachable
        assert False

    async def probe(self) -> Version:
        rsp = await self.send_command(FunctionID.SERIAL_API_GET_CAPABILITIES)

        major = rsp.data[0]
        minor = rsp.data[1]

        return Version(f"{major}.{minor}")

    async def enter_bootloader(self) -> None:
        """Reboot the device into the bootloader. No response is expected."""
        frame = ZWaveFrame(
            type=MessageType.REQUEST,
            function_id=FunctionID.SERIAL_API_ENTER_BOOTLOADER,
            data=b"",
        )
        self.send_data(frame.serialize())



from __future__ import annotations

import asyncio
import logging
from unittest.mock import Mock, patch

import pytest

from universal_silabs_flasher.common import BufferTooShort, Version
from universal_silabs_flasher.zwave import (
    ACK,
    FunctionID,
    MessageType,
    ZWaveFrame,
    ZWaveProtocol,
)

from .common import PairedTransport

ENTER_BOOTLOADER_REQUEST = bytes.fromhex("01030027db")
GET_CAPABILITIES_REQUEST = bytes.fromhex("01030007fb")
GET_CAPABILITIES_RESPONSE = ZWaveFrame(
    type=MessageType.RESPONSE,
    function_id=FunctionID.SERIAL_API_GET_CAPABILITIES,
    data=bytes.fromhex(
        "0102046600010001f6873e88cf2bc05fe3d7fde0970f008000808680ba0d00f000002e7fe0800000"
    ),
)


def test_frame_serialize() -> None:
    frame = ZWaveFrame(
        type=MessageType.REQUEST,
        function_id=FunctionID.SERIAL_API_GET_CAPABILITIES,
        data=b"",
    )
    assert frame.serialize() == GET_CAPABILITIES_REQUEST

    frame = ZWaveFrame(
        type=MessageType.REQUEST,
        function_id=FunctionID.SERIAL_API_ENTER_BOOTLOADER,
        data=b"",
    )
    assert frame.serialize() == ENTER_BOOTLOADER_REQUEST


def test_frame_deserialize_roundtrip() -> None:
    for frame in (
        ZWaveFrame(
            type=MessageType.REQUEST,
            function_id=FunctionID.SERIAL_API_GET_CAPABILITIES,
            data=b"",
        ),
        GET_CAPABILITIES_RESPONSE,
    ):
        parsed, remainder = ZWaveFrame.deserialize(frame.serialize())
        assert parsed == frame
        assert remainder == b""


def test_frame_deserialize_returns_remainder() -> None:
    raw = GET_CAPABILITIES_RESPONSE.serialize() + b"\xab\xcd"
    parsed, remainder = ZWaveFrame.deserialize(raw)
    assert parsed == GET_CAPABILITIES_RESPONSE
    assert remainder == b"\xab\xcd"


def test_frame_deserialize_too_short() -> None:
    with pytest.raises(BufferTooShort):
        ZWaveFrame.deserialize(b"\x01")

    with pytest.raises(BufferTooShort):
        ZWaveFrame.deserialize(b"\x01\x10")  # claims 16 bytes but body is missing


def test_frame_deserialize_bad_checksum() -> None:
    raw = bytearray(GET_CAPABILITIES_REQUEST)
    raw[-1] ^= 0xFF  # corrupt checksum
    with pytest.raises(ValueError, match="Checksum mismatch"):
        ZWaveFrame.deserialize(bytes(raw))


def test_frame_deserialize_wrong_sof() -> None:
    with pytest.raises(ValueError, match="Expected SOF"):
        ZWaveFrame.deserialize(b"\x00\x03\x00\x07\xfb")


@pytest.mark.parametrize(
    "chunks",
    [
        [GET_CAPABILITIES_REQUEST],
        [bytes([b]) for b in GET_CAPABILITIES_REQUEST],
    ],
)
def test_protocol_data_received(chunks: list[bytes]) -> None:
    protocol = ZWaveProtocol()
    protocol.connection_made(Mock())

    with patch.object(protocol, "frame_received") as mock:
        for chunk in chunks:
            protocol.data_received(chunk)

    assert mock.call_count == 1
    parsed_frame = mock.call_args[0][0]
    assert parsed_frame == ZWaveFrame(
        type=MessageType.REQUEST,
        function_id=FunctionID.SERIAL_API_GET_CAPABILITIES,
        data=b"",
    )


def test_protocol_sends_ack_on_valid_frame() -> None:
    loop = asyncio.new_event_loop()
    protocol = ZWaveProtocol()
    received = bytearray()

    class CapturingTransport(asyncio.Transport):
        def write(self, data):
            received.extend(data)

        def is_closing(self):
            return False

    protocol.connection_made(CapturingTransport())
    protocol.data_received(GET_CAPABILITIES_REQUEST)

    assert received == bytes([ACK])
    loop.close()


def test_protocol_skips_ack_nak_can_and_junk() -> None:
    protocol = ZWaveProtocol()
    protocol.connection_made(Mock())

    with patch.object(protocol, "frame_received") as mock:
        # ACK + NAK + CAN + junk + valid frame
        protocol.data_received(
            bytes([0x06, 0x15, 0x18, 0xAA]) + GET_CAPABILITIES_REQUEST
        )

    assert mock.call_count == 1


def test_protocol_bad_checksum_recovers() -> None:
    protocol = ZWaveProtocol()
    protocol.connection_made(Mock())

    bad = bytearray(GET_CAPABILITIES_REQUEST)
    bad[-1] ^= 0xFF

    with patch.object(protocol, "frame_received") as mock:
        protocol.data_received(bytes(bad) + GET_CAPABILITIES_REQUEST)

    assert mock.call_count == 1


class EchoSide(asyncio.Protocol):
    def __init__(self) -> None:
        self.transport: asyncio.Transport | None = None
        self.received = bytearray()

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        self.transport = transport  # type: ignore[assignment]

    def data_received(self, data: bytes) -> None:
        self.received.extend(data)


async def _make_pair() -> tuple[ZWaveProtocol, EchoSide]:
    loop = asyncio.get_running_loop()

    client = ZWaveProtocol()
    server = EchoSide()

    client.connection_made(PairedTransport(server, loop))
    server.connection_made(PairedTransport(client, loop))

    return client, server


async def test_probe() -> None:
    client, server = await _make_pair()

    async def _respond():
        await asyncio.sleep(0.01)
        assert server.transport is not None
        server.transport.write(GET_CAPABILITIES_RESPONSE.serialize())

    asyncio.create_task(_respond())
    version = await client.probe()
    assert version == Version("1.2")


async def test_probe_timeout() -> None:
    client, _ = await _make_pair()

    with pytest.raises(asyncio.TimeoutError):
        await client.send_command(
            FunctionID.SERIAL_API_GET_CAPABILITIES, timeout=0.05, retries=0
        )


async def test_enter_bootloader() -> None:
    client, server = await _make_pair()
    await client.enter_bootloader()
    await asyncio.sleep(0.01)
    assert ENTER_BOOTLOADER_REQUEST in bytes(server.received)


async def test_unsolicited_response_ignored() -> None:
    client, server = await _make_pair()

    # Send an unsolicited response with no pending future — should not raise
    assert server.transport is not None
    server.transport.write(GET_CAPABILITIES_RESPONSE.serialize())
    await asyncio.sleep(0.01)
    assert not client._pending_frames


async def test_duplicate_response_ignored(caplog: pytest.LogCaptureFixture) -> None:
    client, server = await _make_pair()

    async def _respond_twice():
        await asyncio.sleep(0.01)
        assert server.transport is not None
        server.transport.write(GET_CAPABILITIES_RESPONSE.serialize())
        server.transport.write(GET_CAPABILITIES_RESPONSE.serialize())

    asyncio.create_task(_respond_twice())

    with caplog.at_level(logging.DEBUG):
        result = await client.send_command(FunctionID.SERIAL_API_GET_CAPABILITIES)

    assert result == GET_CAPABILITIES_RESPONSE
    assert "duplicate" in caplog.text



from __future__ import annotations

import dataclasses
import json
import logging
import typing

from zigpy.ota.validators import ValidationError, parse_silabs_ebl, parse_silabs_gbl
import zigpy.types as zigpy_t

from .common import Version, pad_to_multiple
from .const import LEGACY_FIRMWARE_TYPE_REMAPPING, FirmwareImageType

_LOGGER = logging.getLogger(__name__)

NABUCASA_METADATA_VERSION = 2


class GBLTagId(zigpy_t.enum32):
    # First tag in the file. The header tag contains the version number of the GBL file
    # specification, and flags indicating the type of GBL file – whether it is signed
    # or encrypted.
    HEADER = 0x03A617EB
    # Information about the application update image that is contained in this GBL
    # file.
    APP_INFO = 0xF40A0AF4
    # A complete encrypted Secure Element update image. Only applicable on Series 2
    # devices.
    SE_UPGRADE = 0x5EA617EB
    # A complete bootloader update image.
    BOOTLOADER = 0xF50909F5
    # Information about what application data to program at a specific address into the
    # main flash memory. The two tags are interchangeable.
    PROGRAM_DATA1 = 0xFE0101FE
    PROGRAM_DATA2 = 0xFD0303FD
    # LZ4 compressed information about what application data to program at a specific
    # address into the main flash memory.
    PROGRAM_DATA_LZ4 = 0xFD0505FD
    # LZMA compressed information about what application data to program at a specific
    # address into the main flash memory.
    PROGRAM_DATA_LZMA = 0xFD0707FD
    # Metadata that the bootloader does not parse, but can be returned to the
    # application through a callback.
    METADATA = 0xF60808F6
    # The ECDSA-P256 signature of all preceding data in the file.
    SIGNATURE = 0xF70A0AF7
    # End of the GBL file. It contains a 32-bit CRC for the entire file as an integrity
    # check. The CRC is a non-cryptographic check. This must be the last tag.
    END = 0xFC0404FC


class EBLTagId(zigpy_t.enum16):
    # TODO: flip the endianness
    HEADER = 0x0000
    PROG = 0x01FE
    MFGPROG = 0xFE02
    ERASEPROG = 0x03FD
    END = 0x04FC
    ENC_HEADER = 0x05FB
    ENC_INIT = 0x06FA
    ENC_EBL_DATA = 0x07F9
    ENC_MAC = 0x09F7


@dataclasses.dataclass(frozen=True)
class NabuCasaMetadata:
    metadata_version: int

    sdk_version: Version | None = None
    ezsp_version: Version | None = None
    ot_rcp_version: Version | None = None
    cpc_version: Version | None = None
    zwave_version: Version | None = None

    fw_type: FirmwareImageType | None = None
    fw_variant: str | None = None
    baudrate: int | None = None

    original_json: dict[str, typing.Any] = dataclasses.field(
        repr=False, default_factory=dict
    )

    def get_public_version(self) -> Version | None:
        return (
            self.cpc_version
            or self.ezsp_version
            or self.ot_rcp_version
            or self.zwave_version
            or self.sdk_version
        )

    @classmethod
    def from_json(cls, obj: dict[str, typing.Any]) -> NabuCasaMetadata:
        original_json = json.loads(json.dumps(obj))
        metadata_version = obj.pop("metadata_version")

        if metadata_version > NABUCASA_METADATA_VERSION:
            raise ValueError(
                f"Unknown metadata version: {metadata_version},"
                f" expected {NABUCASA_METADATA_VERSION}"
            )

        if sdk_version := obj.pop("sdk_version", None):
            sdk_version = Version(sdk_version)

        if ezsp_version := obj.pop("ezsp_version", None):
            ezsp_version = Version(ezsp_version)

        if ot_rcp_version := obj.pop("ot_rcp_version", None):
            ot_rcp_version = Version(ot_rcp_version)

        if cpc_version := obj.pop("cpc_version", None):
            cpc_version = Version(cpc_version)

        if zwave_version := obj.pop("zwave_version", None):
            zwave_version = Version(zwave_version)

        if fw_type := obj.pop("fw_type", None):
            if fw_type in LEGACY_FIRMWARE_TYPE_REMAPPING:
                fw_type = LEGACY_FIRMWARE_TYPE_REMAPPING[fw_type]

            try:
                fw_type = FirmwareImageType(fw_type)
            except ValueError:
                _LOGGER.warning("Unknown firmware type: %r", fw_type)
                fw_type = None

        if fw_variant := obj.pop("fw_variant", None):
            fw_variant = fw_variant

        baudrate = obj.pop("baudrate", None)

        if obj:
            _LOGGER.warning("Unexpected keys in JSON remain: %r", obj)

        return cls(
            metadata_version=metadata_version,
            sdk_version=sdk_version,
            ezsp_version=ezsp_version,
            ot_rcp_version=ot_rcp_version,
            cpc_version=cpc_version,
            zwave_version=zwave_version,
            fw_type=fw_type,
            fw_variant=fw_variant,
            baudrate=baudrate,
            original_json=original_json,
        )


TagId = typing.TypeVar("TagId")


@dataclasses.dataclass(frozen=True)
class FirmwareImage(typing.Generic[TagId]):
    tags: list[tuple[TagId, bytes]]

    @classmethod
    def from_bytes(cls, data: bytes) -> FirmwareImage[typing.Any]:
        raise NotImplementedError()

    def serialize(self) -> bytes:
        raise NotImplementedError()

    def get_nabucasa_metadata(self) -> NabuCasaMetadata:
        raise KeyError("Metadata not available for this firmware type")

    def get_first_tag(self, tag_id: TagId) -> bytes:
        try:
            return next(v for t, v in self.tags if t == tag_id)
        except StopIteration:
            raise KeyError(f"No tag with id {tag_id!r} exists")


@dataclasses.dataclass(frozen=True)
class GBLImage(FirmwareImage[GBLTagId]):
    @classmethod
    def from_bytes(cls, data: bytes | memoryview) -> GBLImage:
        if isinstance(data, memoryview):
            data = data.tobytes()

        tags = []

        for tag_bytes, value in parse_silabs_gbl(data):
            tag, _ = GBLTagId.deserialize(tag_bytes)
            tags.append((tag, value))

        return cls(tags=tags)

    def serialize(self) -> bytes:
        return pad_to_multiple(
            b"".join(
                [
                    tag_id.serialize() + len(value).to_bytes(4, "little") + value
                    for tag_id, value in self.tags
                ]
            ),
            4,
            b"\xff",
        )

    def get_nabucasa_metadata(self) -> NabuCasaMetadata:
        metadata = self.get_first_tag(GBLTagId.METADATA)

        return NabuCasaMetadata.from_json(json.loads(metadata))


@dataclasses.dataclass(frozen=True)
class EBLImage(FirmwareImage[EBLTagId]):
    @classmethod
    def from_bytes(cls, data: bytes) -> EBLImage:
        tags = []

        for tag_bytes, value in parse_silabs_ebl(data):
            tag, _ = EBLTagId.deserialize(tag_bytes)
            tags.append((tag, value))

        return cls(tags=tags)

    def serialize(self) -> bytes:
        return pad_to_multiple(
            b"".join(
                [
                    tag_id.serialize() + len(value).to_bytes(2, "big") + value
                    for tag_id, value in self.tags
                ]
            ),
            64,
            b"\xff",
        )

    def get_nabucasa_metadata(self) -> NabuCasaMetadata:
        raise KeyError("Metadata not supported for EBL")


def parse_firmware_image(data: bytes) -> FirmwareImage[typing.Any]:
    fw_classes: list[type[GBLImage] | type[EBLImage]] = [GBLImage, EBLImage]
    for fw_cls in fw_classes:
        try:
            return fw_cls.from_bytes(data)
        except ValidationError:
            pass

    raise ValueError("Unknown firmware image type")



from __future__ import annotations

import asyncio
import logging
import sys
from unittest.mock import MagicMock, call

import pytest

from universal_silabs_flasher.spinel import (
    HDLCLiteFrame,
    SpinelFrame,
    SpinelHeader,
    SpinelProtocol,
)
from universal_silabs_flasher.spinel_types import CommandID, PackedUInt21, PropertyID

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout  # pragma: no cover
else:
    from asyncio import timeout as asyncio_timeout  # pragma: no cover

from universal_silabs_flasher.common import crc16_kermit
import universal_silabs_flasher.spinel as spinel

from .common import PairedTransport

_LOGGER = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "data, crc",
    [
        (b"", 0x0000),
        (b"foobar", 0x147B),
        (b"\xfa\x9b\x51\xb9\xf2\x53\xe3\xbd", 0x6782),
    ],
)
def test_hdlc_lite_crc(data, crc):
    assert crc16_kermit(data) == crc


@pytest.mark.parametrize(
    "encoded, decoded",
    [
        (bytes.fromhex("7e810243d3d37e"), bytes.fromhex("810243")),
        (bytes.fromhex("7e8103367d5e7d5d6af97e"), bytes.fromhex("8103367e7d")),
        (bytes.fromhex("7e810365010b287e"), bytes.fromhex("81036501")),
        (bytes.fromhex("7e8103862a01547d5e7e"), bytes.fromhex("8103862a01")),
        (
            bytes.fromhex(
                "7e8106024f50454e5448524541442f366666316163302d64697274793b204546523332"
                "3b2044656320323320323032322031383a30383a303000fa8c7e"
            ),
            bytes.fromhex(
                "8106024f50454e5448524541442f366666316163302d64697274793b2045465233323b"
                "2044656320323320323032322031383a30383a303000"
            ),
        ),
    ],
)
def test_hdlc_lite_encoding_decoding(encoded, decoded):
    assert spinel.HDLCLiteFrame(data=decoded).serialize() == encoded
    assert spinel.HDLCLiteFrame.from_bytes(encoded).data == decoded


@pytest.mark.parametrize(
    "encoded, decoded",
    [
        (
            bytes.fromhex(
                "8106024f50454e5448524541442f366666316163302d64697274793b2045465233323b"
                "2044656320323320323032322031383a30383a303000"
            ),
            spinel.SpinelFrame(
                header=spinel.SpinelHeader(
                    transaction_id=1,
                    network_link_id=0,
                    flag=0b10,
                ),
                command_id=CommandID.PROP_VALUE_IS,
                data=b"\x02OPENTHREAD/6ff1ac0-dirty; EFR32; Dec 23 2022 18:08:00\x00",
            ),
        ),
    ],
)
def test_spinel_parsing(encoded, decoded):
    assert spinel.SpinelFrame.from_bytes(encoded) == decoded
    assert decoded.serialize() == encoded


class Conversation(asyncio.Protocol):
    """A helper to script a conversation with a protocol."""

    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
        self.loop = loop
        self.transport: PairedTransport | None = None
        self._reader = asyncio.StreamReader(loop=loop)

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        assert isinstance(transport, PairedTransport)
        self.transport = transport

    def data_received(self, data: bytes) -> None:
        self._reader.feed_data(data)

    def connection_lost(self, exc: Exception | None) -> None:
        self._reader.feed_eof()

    async def send(self, data: bytes) -> None:
        assert self.transport is not None
        self.transport.write(data)


async def create_spinel_test_pair() -> tuple[SpinelProtocol, Conversation]:
    """Creates a connected pair of a client protocol and a conversation helper."""
    loop = asyncio.get_running_loop()
    client = SpinelProtocol()
    conversation = Conversation(loop)

    client_transport = PairedTransport(conversation, loop)
    server_transport = PairedTransport(client, loop)

    client.connection_made(client_transport)
    conversation.connection_made(server_transport)

    return client, conversation


async def test_ignore_duplicate_response(caplog: pytest.CapLogFixture) -> None:
    """Test that duplicate responses are ignored."""
    client, conversation = await create_spinel_test_pair()

    # Send a command and get the future
    fut = asyncio.create_task(client.send_command(CommandID.PROP_VALUE_GET, b"\x01"))
    # Let the command be sent
    await asyncio.sleep(0.01)

    assert len(client._pending_frames) == 1
    tid = list(client._pending_frames.keys())[0]

    # Receive the first response
    response1 = SpinelFrame(
        header=SpinelHeader(transaction_id=tid, network_link_id=0, flag=0b10),
        command_id=CommandID.PROP_VALUE_IS,
        data=b"\x01test",
    )
    await conversation.send(HDLCLiteFrame(data=response1.serialize()).serialize())

    with caplog.at_level(logging.DEBUG):
        # Receive the duplicate response immediately
        response2 = SpinelFrame(
            header=SpinelHeader(transaction_id=tid, network_link_id=0, flag=0b10),
            command_id=CommandID.PROP_VALUE_IS,
            data=b"\x01test2",
        )

        await conversation.send(HDLCLiteFrame(data=response2.serialize()).serialize())
        await asyncio.sleep(0.01)

    # The future should be resolved with the first response
    async with asyncio_timeout(1):
        result = await fut

    assert result == response1
    assert f"Ignoring duplicate response for TID {tid}" in caplog.text


async def test_ignore_tid_zero_response(caplog: pytest.CapLogFixture) -> None:
    """Test that responses with TID 0 are ignored if no future is pending."""
    client, conversation = await create_spinel_test_pair()

    caplog.set_level(logging.DEBUG)
    caplog.clear()

    # We don't have a pending future for TID 0
    assert 0 not in client._pending_frames

    # Receive a response with TID 0
    response = SpinelFrame(
        header=SpinelHeader(transaction_id=0, network_link_id=0, flag=0b10),
        command_id=CommandID.PROP_VALUE_IS,
        data=b"\x01test",
    )
    await conversation.send(HDLCLiteFrame(data=response.serialize()).serialize())
    await asyncio.sleep(0.01)

    # Nothing should have happened
    for record in caplog.records:
        assert record.levelno < logging.WARNING


async def test_property_listener() -> None:
    """Test the property listener mechanism."""
    client, conversation = await create_spinel_test_pair()
    listener = MagicMock()
    prop_id = PropertyID.STREAM_DEBUG
    prop_data = b"hello world"

    client.add_property_listener(prop_id, listener)

    # Send a PROP_VALUE_IS frame
    frame_data = PackedUInt21.serialize(prop_id) + prop_data
    response = SpinelFrame(
        header=SpinelHeader(transaction_id=0, network_link_id=0, flag=0b10),
        command_id=CommandID.PROP_VALUE_IS,
        data=frame_data,
    )
    await conversation.send(HDLCLiteFrame(data=response.serialize()).serialize())
    await asyncio.sleep(0.01)

    listener.mock_calls == [call(prop_data)]

    # Test removing the listener
    listener.reset_mock()
    client.remove_property_listener(prop_id, listener)
    await conversation.send(HDLCLiteFrame(data=response.serialize()).serialize())
    await asyncio.sleep(0.01)

    listener.mock_calls == []


async def test_iter_property_changes() -> None:
    """Test the async iterator for property changes."""
    client, conversation = await create_spinel_test_pair()
    prop_id = PropertyID.STREAM_DEBUG
    queue: asyncio.Queue[bytes] = asyncio.Queue()

    async def _iterator_task():
        async for value in client.iter_property_changes(prop_id):
            await queue.put(value)

    task = asyncio.create_task(_iterator_task())

    # Send a few PROP_VALUE_IS frames
    for i in range(5):
        response = SpinelFrame(
            header=SpinelHeader(transaction_id=0, network_link_id=0, flag=0b10),
            command_id=CommandID.PROP_VALUE_IS,
            data=PackedUInt21.serialize(prop_id) + f"test {i}".encode("ascii"),
        )
        await conversation.send(HDLCLiteFrame(data=response.serialize()).serialize())

    async with asyncio_timeout(1):
        results = [await queue.get() for _ in range(5)]

    assert results == [f"test {i}".encode("ascii") for i in range(5)]

    task.cancel()


async def test_send_frame_tid_zero_raises() -> None:
    """Test that send_frame with tid=0 and wait_response=True raises ValueError."""
    client, _ = await create_spinel_test_pair()
    frame = SpinelFrame(
        header=SpinelHeader(transaction_id=0, network_link_id=0, flag=0b10),
        command_id=CommandID.NOOP,
        data=b"",
    )

    with pytest.raises(ValueError, match="Cannot wait for response on TID=0 frames"):
        await client.send_frame(frame, tid=0, wait_response=True)

    # Should not raise when wait_response is False
    await client.send_frame(frame, tid=0, wait_response=False)


@pytest.mark.parametrize(
    ("value", "serialized"),
    [
        (0, b"\x00"),
        (42, b"\x2a"),
        (200, b"\xc8\x01"),
        (1337, b"\xb9\x0a"),
        (40000, b"\xc0\xb8\x02"),
        (2097151, b"\xff\xff\x7f"),
    ],
)
def test_packed_uint21_serialization(value, serialized):
    """Test PackedUInt21 serialization."""

    assert PackedUInt21(value).serialize() == serialized

    deserialized, remainder = PackedUInt21.deserialize(serialized + b"remainder")
    assert deserialized == value
    assert remainder == b"remainder"


def test_packed_uint21_deserialization_failure():
    """Test PackedUInt21 deserialization failure on oversized input."""

    with pytest.raises(ValueError):
        PackedUInt21.deserialize(b"\x80\x80\x80\x80\x00")



from __future__ import annotations

import asyncio
import dataclasses
import logging
import typing

from zigpy.serial import SerialProtocol
import zigpy.types

from . import cpc_types
from .common import BufferTooShort, Version, asyncio_timeout, crc16_ccitt

_LOGGER = logging.getLogger(__name__)


def parse_subframe(cpc_frame: CPCTransportFrame) -> UnnumberedFrame:
    """Parses a CPC sub-frame from a CPC frame. Only `UnnumberedFrame` is supported."""
    frame_type = cpc_frame.frame_type()

    if frame_type != cpc_types.FrameType.UNNUMBERED:
        raise ValueError(f"Unsupported frame type: {frame_type!r}")

    assert isinstance(cpc_frame.payload, (bytes, bytearray))
    return UnnumberedFrame.from_bytes(cpc_frame.payload)


class Command:
    """Base class for unnumbered commands."""

    @classmethod
    def from_bytes(cls, data: bytes) -> Command:
        raise NotImplementedError

    def to_bytes(self) -> bytes:
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class PropertyCommand(Command):
    """Unnumbered frame command to get/set/read a property."""

    property_id: cpc_types.PropertyId
    value: bytes

    @classmethod
    def from_bytes(cls, data: bytes) -> PropertyCommand:
        property_id, data = cpc_types.PropertyId.deserialize(data)

        return cls(
            property_id=property_id,
            value=data,
        )

    def to_bytes(self) -> bytes:
        return self.property_id.serialize() + self.value


@dataclasses.dataclass(frozen=True)
class ResetCommand(Command):
    """Unnumbered frame command to reset the device."""

    # The `status` field is set when the device responds
    status: cpc_types.Status | None

    @classmethod
    def from_bytes(cls, data: bytes) -> ResetCommand:
        if not data:
            status = None
        else:
            status, rest = cpc_types.Status.deserialize(data)
            assert not rest

        return cls(status=status)

    def to_bytes(self) -> bytes:
        return self.status.serialize() if self.status is not None else b""


@dataclasses.dataclass(frozen=True)
class UnnumberedFrame:
    """Unnumbered CPC frame"""

    command_id: cpc_types.UnnumberedFrameCommandId
    command_seq: zigpy.types.uint8_t
    payload: Command

    _COMMANDS: typing.ClassVar[
        dict[
            cpc_types.UnnumberedFrameCommandId,
            type[PropertyCommand] | type[ResetCommand],
        ]
    ] = {
        cpc_types.UnnumberedFrameCommandId.PROP_VALUE_GET: PropertyCommand,
        cpc_types.UnnumberedFrameCommandId.PROP_VALUE_SET: PropertyCommand,
        cpc_types.UnnumberedFrameCommandId.PROP_VALUE_IS: PropertyCommand,
        cpc_types.UnnumberedFrameCommandId.RESET: ResetCommand,
    }

    @classmethod
    def from_bytes(cls, data: bytes) -> UnnumberedFrame:
        command_id, data = cpc_types.UnnumberedFrameCommandId.deserialize(data)
        command_seq, data = zigpy.types.uint8_t.deserialize(data)
        length, data = zigpy.types.uint16_t.deserialize(data)

        if len(data) < length:
            raise ValueError("Frame is too short")

        payload, data = data[:length], data[length:]

        if data:
            raise ValueError("Trailing data in frame")

        return cls(
            command_id=command_id,
            command_seq=command_seq,
            payload=cls._COMMANDS[command_id].from_bytes(payload),
        )

    def to_bytes(self) -> bytes:
        payload = self.payload.to_bytes()

        return (
            self.command_id.serialize()
            + self.command_seq.serialize()
            + zigpy.types.uint16_t(len(payload)).serialize()  # type: ignore[no-untyped-call]
            + payload
        )


@dataclasses.dataclass(frozen=True)
class CPCTransportFrame:
    """CPC transport frame"""

    endpoint: cpc_types.EndpointId
    control: zigpy.types.uint8_t
    payload: bytes | UnnumberedFrame

    def serialize(self) -> bytes:
        """Serialize the transport frame and compute lengths and checksums."""
        assert isinstance(self.payload, UnnumberedFrame)
        payload = self.payload.to_bytes()
        length = zigpy.types.uint16_t(len(payload) + 2)  # type: ignore[no-untyped-call]

        header = (
            cpc_types.FLAG.serialize()
            + self.endpoint.serialize()
            + length.serialize()
            + self.control.serialize()
        )

        header_checksum = crc16_ccitt(header).to_bytes(2, "little")
        payload_checksum = crc16_ccitt(payload).to_bytes(2, "little")

        return header + header_checksum + payload + payload_checksum

    @classmethod
    def deserialize(cls, data: bytes | bytearray) -> tuple[CPCTransportFrame, bytes]:
        if len(data) < 7:
            raise BufferTooShort("Data is too short to contain packet header")

        orig_data = data
        flag, data = zigpy.types.uint8_t.deserialize(data)  # type: ignore[arg-type]

        if flag != cpc_types.FLAG:
            raise ValueError("Invalid flag")

        endpoint, data = cpc_types.EndpointId.deserialize(data)
        length, data = zigpy.types.uint16_t.deserialize(data)
        control, data = zigpy.types.uint8_t.deserialize(data)
        header_checksum, data = zigpy.types.uint16_t.deserialize(data)

        if crc16_ccitt(orig_data[:5]) != header_checksum:
            raise ValueError("Invalid header checksum")

        if len(data) < length:
            raise BufferTooShort("Data is too short to contain packet payload")

        payload, data = data[: length - 2], data[length - 2 :]
        payload_checksum, data = zigpy.types.uint16_t.deserialize(data)

        if crc16_ccitt(payload) != payload_checksum:
            raise ValueError("Invalid payload checksum")

        frame = cls(
            endpoint=endpoint,
            control=control,
            payload=payload,
        )

        frame_with_parsed_payload = dataclasses.replace(
            frame, payload=parse_subframe(frame)
        )

        return frame_with_parsed_payload, data

    def frame_type(self) -> cpc_types.FrameType:
        frame_type = (self.control & 0b11000000) >> 6

        if frame_type == 0:
            frame_type = 1

        return cpc_types.FrameType(frame_type)  # type: ignore[no-untyped-call]

    def seq(self) -> int:
        return (self.control & 0b01110000) >> 4

    def supervisory_function(self) -> int:
        assert self.frame_type() == cpc_types.FrameType.SUPERVISORY
        return (self.control & 0b00110000) >> 4

    def ack(self) -> int:
        return (self.control & 0b00000111) >> 0

    def unnumbered_type(self) -> cpc_types.UnnumberedFrameType:
        assert self.frame_type() == cpc_types.FrameType.UNNUMBERED
        return cpc_types.UnnumberedFrameType((self.control & 0b00111111) >> 0)  # type: ignore[no-untyped-call]

    def poll_final(self) -> bool:
        return bool((self.control & 0b00001000) >> 3)


class CPCProtocol(SerialProtocol):
    """Partial implementation of the CPC protocol."""

    _buffer: bytearray

    def __init__(self) -> None:
        super().__init__()
        self._command_seq: int = 0
        self._pending_frames: dict[int, asyncio.Future[CPCTransportFrame]] = {}

    async def probe(self) -> Version:
        cpc_version = await self.get_cpc_version()
        secondary_version = await self.get_secondary_version()

        # Prefer the secondary version if possible, on newer firmwares we customize it
        if secondary_version is not None:
            return secondary_version

        return cpc_version

    async def enter_bootloader(self) -> None:
        """Reboot into the bootloader."""
        await self.send_unnumbered_frame(
            command_id=cpc_types.UnnumberedFrameCommandId.PROP_VALUE_SET,
            command_payload=PropertyCommand(
                property_id=cpc_types.PropertyId.BOOTLOADER_REBOOT_MODE,
                value=cpc_types.RebootMode.BOOTLOADER.serialize(),
            ),
        )

        await self.send_unnumbered_frame(
            command_id=cpc_types.UnnumberedFrameCommandId.RESET,
            command_payload=ResetCommand(status=None),
        )

        # A small delay is necessary when switching baudrates
        await asyncio.sleep(0.5)

    async def get_cpc_version(self) -> Version:
        """Read the secondary CPC version from the device."""
        rsp = await self.send_unnumbered_frame(
            command_id=cpc_types.UnnumberedFrameCommandId.PROP_VALUE_GET,
            command_payload=PropertyCommand(
                property_id=cpc_types.PropertyId.SECONDARY_CPC_VERSION,
                value=b"",
            ),
            retries=3,
        )

        assert isinstance(rsp.payload, UnnumberedFrame)
        assert isinstance(rsp.payload.payload, PropertyCommand)
        version_bytes = rsp.payload.payload.value
        major, version_bytes = zigpy.types.uint32_t.deserialize(version_bytes)
        minor, version_bytes = zigpy.types.uint32_t.deserialize(version_bytes)
        patch, version_bytes = zigpy.types.uint32_t.deserialize(version_bytes)
        assert not version_bytes

        return Version(f"{major}.{minor}.{patch}")

    async def get_secondary_version(self) -> Version | None:
        """Read the secondary app version from the device."""
        rsp = await self.send_unnumbered_frame(
            command_id=cpc_types.UnnumberedFrameCommandId.PROP_VALUE_GET,
            command_payload=PropertyCommand(
                property_id=cpc_types.PropertyId.SECONDARY_APP_VERSION,
                value=b"",
            ),
            retries=3,
        )

        assert isinstance(rsp.payload, UnnumberedFrame)
        assert isinstance(rsp.payload.payload, PropertyCommand)
        version_bytes = rsp.payload.payload.value

        if version_bytes == b"UNDEFINED\x00":
            return None

        return Version(version_bytes.split(b"\x00", 1)[0].decode("ascii"))

    def send_data(self, data: bytes) -> None:
        assert self._transport is not None
        _LOGGER.debug("Sending data %s", data)
        self._transport.write(data)

    def data_received(self, data: bytes) -> None:
        super().data_received(data)

        while self._buffer:
            try:
                frame, new_buffer = CPCTransportFrame.deserialize(self._buffer)
                self._buffer = typing.cast(bytearray, new_buffer)
            except BufferTooShort:
                break
            except ValueError as e:
                _LOGGER.debug("Failed to parse buffer %r: %r", self._buffer, e)
                flag = bytes([cpc_types.FLAG])

                try:
                    flag_index = self._buffer.index(flag, 1)
                except ValueError:
                    self._buffer.clear()
                    break
                else:
                    self._buffer = self._buffer[flag_index:]
            else:
                self.frame_received(frame)

    def frame_received(self, frame: CPCTransportFrame) -> None:
        _LOGGER.debug("Parsed frame %s %s", frame.unnumbered_type(), frame)

        if frame.unnumbered_type() == cpc_types.UnnumberedFrameType.POLL_FINAL:
            assert isinstance(frame.payload, UnnumberedFrame)
            if frame.payload.command_seq not in self._pending_frames:
                _LOGGER.debug("Received an unsolicited frame: %s", frame)
                return

            future = self._pending_frames.pop(frame.payload.command_seq)
            future.set_result(frame)

    async def send_unnumbered_frame(
        self,
        command_id: cpc_types.UnnumberedFrameCommandId,
        command_payload: Command,
        *,
        retries: int = 3,
        timeout: float = 1,
        retry_delay: float = 0.1,
    ) -> CPCTransportFrame:
        """Send an unnumbered frame to the device and return the response."""
        unnumbered = UnnumberedFrame(
            command_id=command_id,
            command_seq=zigpy.types.uint8_t(self._command_seq),  # type: ignore[no-untyped-call]
            payload=command_payload,
        )
        frame = CPCTransportFrame(
            endpoint=cpc_types.EndpointId.SYSTEM,
            control=zigpy.types.uint8_t(  # type: ignore[no-untyped-call]
                (cpc_types.FrameType.UNNUMBERED << 6)
                | (cpc_types.UnnumberedFrameType.POLL_FINAL << 0)
            ),
            payload=unnumbered,
        )
        self._command_seq = (self._command_seq + 1) & 0xFF

        assert self._command_seq not in self._pending_frames

        future = asyncio.get_running_loop().create_future()
        self._pending_frames[unnumbered.command_seq] = future

        try:
            for attempt in range(retries + 1):
                _LOGGER.debug("Sending frame %s", frame)
                self.send_data(frame.serialize())

                try:
                    async with asyncio_timeout(timeout):
                        return await asyncio.shield(future)
                except asyncio.TimeoutError:
                    _LOGGER.debug(
                        "Failed to send %s, trying again in %0.2fs (attempt %s of %s)",
                        frame,
                        retry_delay,
                        attempt + 1,
                        retries + 1,
                    )

                    if attempt >= retries:
                        raise

                    await asyncio.sleep(retry_delay)
        finally:
            self._pending_frames.pop(unnumbered.command_seq, None)

        raise AssertionError("Unreachable")



import asyncio
from collections.abc import Generator
from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest
import zigpy.types as t

from universal_silabs_flasher.common import FlowControlSerialProtocol, Version
from universal_silabs_flasher.const import (
    RESET_CONFIGS,
    ApplicationType,
    BaudrateResetConfig,
    GpioPattern,
    GpioResetConfig,
    ModemPinPattern,
    ModemPinResetConfig,
    ResetTarget,
)
from universal_silabs_flasher.flasher import (
    Flasher,
    ProbeResult,
    YellowFlasher,
    Zbt1Flasher,
    Zbt2Flasher,
)
from universal_silabs_flasher.gecko_bootloader import GeckoBootloaderProtocol


@pytest.fixture(autouse=True)
def reduce_timeouts() -> Generator[None, None, None]:
    with patch(
        "universal_silabs_flasher.flasher.BaseFlasher._bootloader_launch_delay", 0.05
    ):
        yield


async def test_write_emberznet_eui64():
    flasher = Flasher(device="/dev/ttyMOCK")

    with (
        patch.object(
            flasher, "probe_gecko_bootloader", side_effect=asyncio.TimeoutError
        ),
        patch.object(
            flasher,
            "probe_ezsp",
            return_value=ProbeResult(
                version=Version("7.4.4.0 build 0"),
                continue_probing=False,
                baudrate=115200,
            ),
        ),
        patch.object(flasher, "_connect_ezsp") as mock_connect_ezsp,
    ):
        ezsp = mock_connect_ezsp.return_value.__aenter__.return_value

        ezsp.getEui64.return_value = (t.EUI64.convert("00:11:22:33:44:55:66:77"),)
        ezsp.write_custom_eui64.return_value = None

        await flasher.write_emberznet_eui64(
            new_ieee=t.EUI64.convert("11:22:33:44:55:66:77:88"), force=True
        )

    assert ezsp.write_custom_eui64.mock_calls == [
        call(ieee=t.EUI64.convert("11:22:33:44:55:66:77:88"), burn_into_userdata=True)
    ]


async def test_baudrate_reset_pattern():
    flasher = Flasher(device="/dev/ttyMOCK", bootloader_reset=(ResetTarget.BAUDRATE,))

    with patch(
        "universal_silabs_flasher.flasher.connect_protocol"
    ) as mock_connect_protocol:
        mock_uart = mock_connect_protocol.return_value.__aenter__.return_value
        mock_uart._transport.write = MagicMock()
        await flasher.trigger_bootloader_reset(run_firmware=False)

    assert mock_connect_protocol.mock_calls == [
        # Connect with 150 baud
        call("/dev/ttyMOCK", 150, FlowControlSerialProtocol),
        call().__aenter__(),
        call().__aexit__(None, None, None),
        # Connect with 300 baud
        call("/dev/ttyMOCK", 300, FlowControlSerialProtocol),
        call().__aenter__(),
        call().__aexit__(None, None, None),
        # Connect with 1200 baud
        call("/dev/ttyMOCK", 1200, FlowControlSerialProtocol),
        call().__aenter__(),
        call().__aenter__()._transport.write(b"BZ"),
        call().__aexit__(None, None, None),
        # Probe
        call("/dev/ttyMOCK", 115200, GeckoBootloaderProtocol),
        call().__aenter__(),
        call().__aenter__().probe(),
        call().__aexit__(None, None, None),
    ]


async def test_trigger_bootloader_reset_first_probe_succeeds():
    flasher = Flasher(
        device="/dev/ttyMOCK",
        bootloader_reset=(ResetTarget.RTS_DTR, ResetTarget.BAUDRATE),
    )

    with (
        patch.object(flasher, "_trigger_baudrate_reset") as mock_trigger_baudrate,
        patch.object(flasher, "_trigger_modem_pin_reset") as mock_trigger_modem_pin,
        patch.object(
            flasher,
            "probe_gecko_bootloader",
            return_value=ProbeResult(
                version=Version("1.0.0"),
                continue_probing=False,
                baudrate=115200,
            ),
        ) as mock_probe,
    ):
        result = await flasher.trigger_bootloader_reset(run_firmware=False)

    assert result is not None
    assert result.version == Version("1.0.0")
    assert result.baudrate == 115200

    # All reset targets are triggered upfront
    assert len(mock_trigger_baudrate.mock_calls) == 1
    assert len(mock_trigger_modem_pin.mock_calls) == 1
    # Only one probe attempt since the first one succeeds
    assert mock_probe.mock_calls == [call(run_firmware=False, baudrate=115200)]


async def test_trigger_bootloader_reset_all_probes_fail():
    flasher = Flasher(
        device="/dev/ttyMOCK",
        bootloader_reset=(ResetTarget.RTS_DTR, ResetTarget.BAUDRATE),
    )

    with (
        patch.object(flasher, "_trigger_baudrate_reset") as mock_trigger_baudrate,
        patch.object(flasher, "_trigger_modem_pin_reset") as mock_trigger_modem_pin,
        patch.object(
            flasher,
            "probe_gecko_bootloader",
            side_effect=asyncio.TimeoutError,  # All probes fail
        ) as mock_probe,
    ):
        result = await flasher.trigger_bootloader_reset(run_firmware=False)

    assert result is None

    # All reset targets are triggered upfront
    assert len(mock_trigger_baudrate.mock_calls) == 1
    assert len(mock_trigger_modem_pin.mock_calls) == 1
    # One probe attempt at the only bootloader baudrate
    assert mock_probe.mock_calls == [
        call(run_firmware=False, baudrate=115200),
    ]


async def test_probe_app_type_fallback_to_bootloader() -> None:
    flasher = Flasher(device="/dev/ttyMOCK", bootloader_reset=(ResetTarget.RTS_DTR,))

    bootloader_result = ProbeResult(
        version=Version("1.0.0"),
        continue_probing=False,
        baudrate=115200,
    )

    with (
        patch.object(
            flasher, "trigger_bootloader_reset", return_value=bootloader_result
        ) as mock_trigger_reset,
        patch.object(
            flasher, "probe_gecko_bootloader", side_effect=asyncio.TimeoutError
        ),
        patch.object(flasher, "probe_cpc", side_effect=asyncio.TimeoutError),
        patch.object(flasher, "probe_ezsp", side_effect=asyncio.TimeoutError),
        patch.object(flasher, "probe_router", side_effect=asyncio.TimeoutError),
        patch.object(flasher, "probe_spinel", side_effect=asyncio.TimeoutError),
    ):
        await flasher.probe_app_type()

    # Should fallback to bootloader when no valid application found
    assert flasher.app_type == ApplicationType.GECKO_BOOTLOADER
    assert flasher.app_version == Version("1.0.0")
    assert flasher.app_baudrate == 115200
    assert flasher.bootloader_baudrate == 115200

    # trigger_bootloader_reset should be called twice - once at start and once
    # for fallback
    assert mock_trigger_reset.call_count == 2


async def test_device_specific_probe_app_type_does_not_require_reset_targets() -> None:
    flasher = Zbt2Flasher(device="/dev/ttyMOCK")

    bootloader_result = ProbeResult(
        version=Version("1.0.0"),
        continue_probing=False,
        baudrate=115200,
    )

    with patch.object(
        flasher,
        "trigger_bootloader_reset",
        side_effect=[bootloader_result, bootloader_result],
    ) as mock_trigger_reset:
        await flasher.probe_app_type()

    assert flasher.app_type == ApplicationType.GECKO_BOOTLOADER
    assert flasher.app_version == Version("1.0.0")
    assert flasher.app_baudrate == 115200
    assert flasher.bootloader_baudrate == 115200
    assert mock_trigger_reset.mock_calls == [
        call(run_firmware=True),
        call(run_firmware=False),
    ]


async def test_flasher_init_string_bootloader_reset():
    flasher = Flasher(device="/dev/ttyMOCK", bootloader_reset="yellow")

    assert flasher._reset_targets == [ResetTarget.YELLOW]


async def test_trigger_gpio_reset_cp210x():
    flasher = Flasher(device="/dev/ttyMOCK")

    config = GpioResetConfig(
        chip=None,
        chip_type="cp210x",
        pattern=[GpioPattern(pins={4: True, 5: True}, delay_after=0.1)],
    )

    with (
        patch(
            "universal_silabs_flasher.flasher.find_gpiochip_by_label",
            return_value="/dev/gpiochip_mock",
        ) as mock_find,
        patch("universal_silabs_flasher.flasher.send_gpio_pattern") as mock_send,
    ):
        await flasher._trigger_gpio_reset(config)

    assert mock_find.mock_calls == [call("cp210x")]
    assert mock_send.mock_calls == [call("/dev/gpiochip_mock", config.pattern)]


async def test_trigger_modem_pin_reset():
    flasher = Flasher(device="/dev/ttyMOCK")

    config = ModemPinResetConfig(
        pattern=[
            ModemPinPattern(pins={"dtr": False, "rts": True}, delay_after=0.0),
            ModemPinPattern(pins={"dtr": True, "rts": False}, delay_after=0.0),
        ],
    )

    with patch(
        "universal_silabs_flasher.flasher.connect_protocol"
    ) as mock_connect_protocol:
        mock_uart = mock_connect_protocol.return_value.__aenter__.return_value
        mock_uart.set_signals = AsyncMock()
        await flasher._trigger_modem_pin_reset(config)

    assert mock_connect_protocol.mock_calls[0] == call(
        "/dev/ttyMOCK", 115200, FlowControlSerialProtocol
    )
    assert mock_uart.set_signals.mock_calls == [
        call(dtr=False, rts=True),
        call(dtr=True, rts=False),
    ]


async def test_reset_config_flasher_trigger_bootloader_reset():
    flasher = YellowFlasher(device="/dev/ttyMOCK")

    assert flasher._can_trigger_bootloader_reset() is True

    probe_result = ProbeResult(
        version=Version("1.0.0"), continue_probing=False, baudrate=115200
    )

    with (
        patch.object(flasher, "_trigger_gpio_reset") as mock_gpio_reset,
        patch.object(
            flasher, "_detect_gecko_bootloader", return_value=probe_result
        ) as mock_detect,
    ):
        result = await flasher.trigger_bootloader_reset(run_firmware=False)

    assert result == probe_result
    assert len(mock_gpio_reset.mock_calls) == 1
    assert mock_detect.mock_calls == [call(run_firmware=False)]


async def test_zbt1_flasher():
    flasher = Zbt1Flasher(device="/dev/ttyMOCK")

    assert flasher._can_trigger_bootloader_reset() is False

    result = await flasher.trigger_bootloader_reset(run_firmware=False)
    assert result is None


async def test_zbt2_flasher_trigger_bootloader_reset_first_attempt_succeeds():
    flasher = Zbt2Flasher(device="/dev/ttyMOCK")

    probe_result = ProbeResult(
        version=Version("1.0.0"), continue_probing=False, baudrate=115200
    )

    with (
        patch.object(flasher, "_trigger_modem_pin_reset") as mock_modem_pin_reset,
        patch.object(flasher, "_trigger_baudrate_reset") as mock_baudrate_reset,
        patch.object(
            flasher, "_detect_gecko_bootloader", return_value=probe_result
        ) as mock_detect,
    ):
        result = await flasher.trigger_bootloader_reset(run_firmware=False)

    assert result == probe_result
    assert mock_modem_pin_reset.mock_calls == [call(RESET_CONFIGS[ResetTarget.RTS_DTR])]
    assert mock_baudrate_reset.mock_calls == [
        call(
            BaudrateResetConfig(
                baudrates=(150, 300, 1200),
                delay_after_each=0.1,
                delay_after_final=0.5,
                command=b"BZ",
            )
        )
    ]
    assert mock_detect.mock_calls == [call(run_firmware=False)]


async def test_zbt2_flasher_trigger_bootloader_reset_hard_reset_fallback():
    flasher = Zbt2Flasher(device="/dev/ttyMOCK")
    flasher._reconnect_timeout = 5

    probe_result = ProbeResult(
        version=Version("1.0.0"), continue_probing=False, baudrate=115200
    )

    with (
        patch.object(flasher, "_trigger_modem_pin_reset") as mock_modem_pin_reset,
        patch.object(
            flasher,
            "_trigger_baudrate_reset",
            side_effect=[
                # Call 1: BZ (initial) - succeeds
                None,
                # Call 2: RE (hard reset) - raises (device disconnects)
                ConnectionError("device disconnected"),
                # Call 3: BZ (reconnect attempt 1) - raises (not ready yet)
                ConnectionError("not ready yet"),
                # Call 4: BZ (reconnect attempt 2) - succeeds
                None,
            ],
        ) as mock_baudrate_reset,
        patch.object(
            flasher,
            "_detect_gecko_bootloader",
            side_effect=[None, probe_result],
        ) as mock_detect,
    ):
        result = await flasher.trigger_bootloader_reset(run_firmware=False)

    assert result == probe_result
    # RTS/DTR reset is only attempted once (before the first BZ)
    assert mock_modem_pin_reset.mock_calls == [call(RESET_CONFIGS[ResetTarget.RTS_DTR])]
    # 4 baudrate reset calls: BZ, RE(fail), BZ(fail), BZ(success)
    assert len(mock_baudrate_reset.mock_calls) == 4
    # 2 detect calls: first returns None, second returns result
    assert len(mock_detect.mock_calls) == 2


async def test_enter_bootloader_from_application():
    flasher = Flasher(device="/dev/ttyMOCK")
    flasher.app_type = ApplicationType.CPC
    flasher.app_baudrate = 115200

    probe_result = ProbeResult(
        version=Version("1.0.0"), continue_probing=False, baudrate=115200
    )

    with (
        patch.object(flasher, "trigger_bootloader_reset", return_value=None),
        patch.object(flasher, "_connect_cpc") as mock_connect_cpc,
        patch.object(flasher, "_detect_gecko_bootloader", return_value=probe_result),
    ):
        mock_cpc = mock_connect_cpc.return_value.__aenter__.return_value
        mock_cpc.enter_bootloader = AsyncMock()
        await flasher.enter_bootloader()

    assert flasher.bootloader_baudrate == 115200
    assert len(mock_cpc.enter_bootloader.mock_calls) == 1



from __future__ import annotations

import asyncio
import collections
import contextlib
import dataclasses
import functools
import logging
import re
import sys
import typing

import crc
import zigpy.serial

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout  # pragma: no cover
else:
    from asyncio import timeout as asyncio_timeout  # pragma: no cover

if typing.TYPE_CHECKING:
    from typing_extensions import Self

__all__ = ["asyncio_timeout"]

_LOGGER = logging.getLogger(__name__)

CONNECT_TIMEOUT = 1
PROBE_TIMEOUT = 2


CRC_CCITT = crc.Calculator(
    crc.Configuration(
        width=16,
        polynomial=0x1021,
        init_value=0x0000,
        final_xor_value=0x0000,
        reverse_input=False,
        reverse_output=False,
    )
)

CRC_KERMIT = crc.Calculator(
    crc.Configuration(
        width=16,
        polynomial=0x1021,
        init_value=0xFFFF,
        final_xor_value=0xFFFF,
        reverse_input=True,
        reverse_output=True,
    )
)


# Used by both CPC and XModem
def crc16_ccitt(data: bytes | bytearray) -> int:
    return int(CRC_CCITT.checksum(data))


# Used by HDLC-Lite
def crc16_kermit(data: bytes | bytearray) -> int:
    return int(CRC_KERMIT.checksum(data))


def pad_to_multiple(data: bytes, multiple: int, padding: bytes) -> bytes:
    assert len(padding) == 1

    if len(data) % multiple == 0:
        return data

    num_complete_blocks = len(data) // multiple
    padded_size = multiple * (num_complete_blocks + 1)

    return data + padding * (padded_size - len(data))


class BufferTooShort(Exception):
    """Protocol buffer requires more data to parse a packet."""


class StateMachine:
    """Asyncio-friendly state machine."""

    def __init__(self, states: set[str], initial: str) -> None:
        if initial not in states:
            raise ValueError(f"Unknown initial state {initial!r}: expected {states!r}")

        self._states = states
        self._state = initial

        self._futures_for_state: collections.defaultdict[
            str, list[asyncio.Future[None]]
        ] = collections.defaultdict(list)

    @property
    def state(self) -> str:
        return self._state

    @state.setter
    def state(self, state: str) -> None:
        if state not in self._states:
            raise ValueError(f"Unknown state {state!r}: expected {self._states!r}")

        self._state = state

        for future in self._futures_for_state[state]:
            future.set_result(None)

    def cancel_all_futures(self, exc: BaseException) -> None:
        for futures in self._futures_for_state.values():
            for future in futures:
                future.set_exception(exc)

    async def wait_for_state(self, state: str, timeout: float = 30.0) -> None:
        """Waits for a state. Returns immediately if the state is active."""
        assert state in self._states

        if self.state == state:
            return

        future: asyncio.Future[None] = asyncio.get_running_loop().create_future()
        self._futures_for_state[state].append(future)

        try:
            async with asyncio_timeout(timeout):
                return await future
        finally:
            # Always clean up the future
            self._futures_for_state[state].remove(future)


P = typing.TypeVar("P", bound=zigpy.serial.SerialProtocol)


@contextlib.asynccontextmanager
async def connect_protocol(
    port: str,
    baudrate: int,
    factory: type[P],
) -> typing.AsyncIterator[P]:
    loop = asyncio.get_running_loop()

    async with asyncio_timeout(CONNECT_TIMEOUT):
        _, protocol = await zigpy.serial.create_serial_connection(
            loop=loop,  # type: ignore[arg-type]  # zigpy expects BaseEventLoop
            protocol_factory=factory,
            url=port,
            baudrate=baudrate,
        )
        serial_protocol = typing.cast(P, protocol)
        await serial_protocol.wait_until_connected()

    try:
        yield serial_protocol
    finally:
        await serial_protocol.disconnect()


def put_first(lst: list[typing.Any], elements: list[typing.Any]) -> list[typing.Any]:
    """Orders a list so that the provided element is first."""
    return elements + [e for e in lst if e not in elements]


@dataclasses.dataclass(frozen=True, order=True)
class VersionComponent:
    comparable: bool
    data: str | int


@functools.total_ordering
class Version:
    _SEPARATORS = {".", "-", "/", "_", " build ", " GA build "}
    _SEPARATORS_REGEX = re.compile(
        "(" + "|".join(re.escape(s) for s in _SEPARATORS) + ")"
    )

    def __init__(self, version: str) -> None:
        self.orig_version = version
        self.components: list[VersionComponent] = []

        # 2.00.01
        # 7.2.2.0 build 190
        # 4.2.2
        # SL-OPENTHREAD/2.2.2.0_GitHub-91fa1f455
        # 4.4.0-2546d625-dirty-676fdb09
        for component in self._SEPARATORS_REGEX.split(version):
            if component.isdigit():
                self.components.append(
                    VersionComponent(comparable=True, data=int(component))
                )
            else:
                self.components.append(
                    VersionComponent(comparable=False, data=component)
                )

    def comparable_components(self) -> tuple[VersionComponent, ...]:
        return tuple(c for c in self.components if c.comparable)

    def compatible_with(self, other: Self) -> bool:
        our_comparable = self.comparable_components()
        their_comparable = other.comparable_components()

        prefix_length = min(len(our_comparable), len(their_comparable))
        return our_comparable[:prefix_length] == their_comparable[:prefix_length]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Version):
            return NotImplemented

        return self.components == other.components

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Version):
            return NotImplemented

        our_comparable = self.comparable_components()
        their_comparable = other.comparable_components()

        return our_comparable < their_comparable

    def __repr__(self) -> str:
        concatenated = "".join(str(c.data) for c in self.components)
        comparable = ".".join(str(c.data) for c in self.comparable_components())

        if concatenated == comparable:
            return f"{concatenated!r}"

        return f"{concatenated!r} ({comparable})"


class FlowControlSerialProtocol(zigpy.serial.SerialProtocol):
    def _set_signals(
        self,
        *,
        rts: bool | None = None,
        cts: bool | None = None,
        dtr: bool | None = None,
    ) -> None:
        assert self._transport is not None
        assert self._transport.serial is not None

        if rts is not None:
            self._transport.serial.rts = rts

        if cts is not None:
            self._transport.serial.cts = cts

        if dtr is not None:
            self._transport.serial.dtr = dtr

    async def set_signals(
        self,
        *,
        rts: bool | None = None,
        cts: bool | None = None,
        dtr: bool | None = None,
    ) -> None:
        assert self._transport is not None

        _LOGGER.debug(
            "Setting UART signals: rts=%s, cts=%s, dtr=%s",
            int(rts) if rts is not None else "-",
            int(cts) if cts is not None else "-",
            int(dtr) if dtr is not None else "-",
        )

        if hasattr(self._transport, "set_signals"):
            await self._transport.set_signals(rts=rts, cts=cts, dtr=dtr)
            return

        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None, lambda: self._set_signals(rts=rts, cts=cts, dtr=dtr)
        )



from argparse import ArgumentTypeError

import pytest

from universal_silabs_flasher.const import ResetTarget
from universal_silabs_flasher.flash import parse_reset_methods


def test_enum_with_separator_single_value() -> None:
    result = parse_reset_methods("rts_dtr")
    assert result == [ResetTarget.RTS_DTR]


def test_enum_with_separator_multiple_values() -> None:
    result = parse_reset_methods("rts_dtr,baudrate")
    assert result == [ResetTarget.RTS_DTR, ResetTarget.BAUDRATE]


def test_enum_with_separator_invalid_value() -> None:
    with pytest.raises(ArgumentTypeError) as exc_info:
        parse_reset_methods("invalid_target")

    assert "'invalid_target' is invalid, must be one of:" in str(exc_info.value)
    assert "yellow" in str(exc_info.value)
    assert "rts_dtr" in str(exc_info.value)



import pathlib

import pytest

from universal_silabs_flasher import firmware
from universal_silabs_flasher.common import Version
from universal_silabs_flasher.const import FirmwareImageType

FIRMWARES_DIR = pathlib.Path(__file__).parent / "firmwares"


def test_firmware_ebl_valid():
    data = (FIRMWARES_DIR / "ncp-uart-sw-6.4.1.ebl").read_bytes()
    fw = firmware.parse_firmware_image(data)

    assert isinstance(fw, firmware.EBLImage)
    assert fw.serialize() == data

    with pytest.raises(KeyError):
        fw.get_nabucasa_metadata()


def test_firmware_gbl_valid_no_metadata():
    data = (
        FIRMWARES_DIR / "NabuCasa_EZSP_v6.10.3.0_PB32_ncp-uart-hw_115200.gbl"
    ).read_bytes()
    fw = firmware.parse_firmware_image(data)

    assert isinstance(fw, firmware.GBLImage)
    assert fw.serialize() == data

    with pytest.raises(KeyError):
        fw.get_nabucasa_metadata()


def test_firmware_gbl_valid_with_metadata():
    data = (
        FIRMWARES_DIR / "NabuCasa_SkyConnect_RCP_v4.1.3_rcp-uart-hw-802154_115200.gbl"
    ).read_bytes()
    fw = firmware.parse_firmware_image(data)

    assert isinstance(fw, firmware.GBLImage)
    assert fw.serialize() == data
    assert fw.get_nabucasa_metadata() == firmware.NabuCasaMetadata(
        metadata_version=1,
        sdk_version=Version("4.1.3"),
        ezsp_version=None,
        cpc_version=None,
        zwave_version=None,
        fw_type=FirmwareImageType.MULTIPAN,
        fw_variant=None,
        ot_rcp_version=None,
        baudrate=None,
        original_json={
            "metadata_version": 1,
            "sdk_version": "4.1.3",
            "fw_type": "rcp-uart-802154",
        },
    )


def test_firmware_gbl_valid_with_metadata_v2():
    data = (FIRMWARES_DIR / "skyconnect_zigbee_ncp_7.4.4.0.gbl").read_bytes()
    fw = firmware.parse_firmware_image(data)

    assert isinstance(fw, firmware.GBLImage)
    assert fw.serialize() == data
    assert fw.get_nabucasa_metadata() == firmware.NabuCasaMetadata(
        metadata_version=2,
        sdk_version=Version("4.4.4"),
        ezsp_version=Version("7.4.4.0"),
        cpc_version=None,
        zwave_version=None,
        fw_type=FirmwareImageType.ZIGBEE_NCP,
        fw_variant=None,
        ot_rcp_version=None,
        baudrate=115200,
        original_json={
            "baudrate": 115200,
            "ezsp_version": "7.4.4.0",
            "fw_type": "zigbee_ncp",
            "fw_variant": None,
            "metadata_version": 2,
            "sdk_version": "4.4.4",
        },
    )



"""CLI integration tests to ensure argument parsing works correctly."""

from dataclasses import dataclass
import io
from pathlib import Path
from unittest.mock import AsyncMock, patch
import zipfile

from aioresponses import aioresponses
import pytest

from universal_silabs_flasher.const import (
    DEFAULT_PROBE_METHODS,
    ApplicationType,
    ResetTarget,
)
from universal_silabs_flasher.flash import main
from universal_silabs_flasher.flasher import BaseFlasher, Flasher, Zbt2Flasher

FIRMWARE_URL = "https://example.com/firmware/skyconnect_zigbee_ncp_7.4.4.0.gbl"
FIRMWARE_PATH = Path("tests/firmwares/skyconnect_zigbee_ncp_7.4.4.0.gbl")


@dataclass
class Result:
    exit_code: str | int
    output: str
    stderr: str
    flasher: BaseFlasher | None


async def invoke_main(argv: list[str]) -> Result:
    """Invoke main() with the given argv, capturing stdout/stderr."""
    stdout = io.StringIO()
    stderr = io.StringIO()

    exit_code: str | int = 0
    captured_flasher = None

    original_init = BaseFlasher.__init__

    def capture_init(self, **kwargs):
        nonlocal captured_flasher
        original_init(self, **kwargs)
        captured_flasher = self

    async def mock_probe_app_type(self):
        self.app_type = ApplicationType.EZSP
        self.app_version = None

    try:
        with (
            patch(
                "universal_silabs_flasher.flasher.BaseFlasher.__init__", capture_init
            ),
            patch("sys.stdout", stdout),
            patch("sys.stderr", stderr),
            patch(
                "universal_silabs_flasher.flasher.BaseFlasher.probe_app_type",
                mock_probe_app_type,
            ),
            patch(
                "universal_silabs_flasher.flasher.Flasher.dump_emberznet_config",
                new_callable=AsyncMock,
            ),
            patch(
                "universal_silabs_flasher.flasher.BaseFlasher.enter_bootloader",
                new_callable=AsyncMock,
            ),
            patch(
                "universal_silabs_flasher.flasher.BaseFlasher.flash_firmware",
                new_callable=AsyncMock,
            ),
        ):
            await main(argv)
    except SystemExit as e:
        exit_code = e.code if e.code is not None else 0

    return Result(
        exit_code=exit_code,
        output=stdout.getvalue(),
        stderr=stderr.getvalue(),
        flasher=captured_flasher,
    )


@pytest.mark.parametrize(
    "args,expected_device,expected_probe_methods,expected_reset",
    [
        # Basic flash command (uses defaults)
        (
            [
                "--device",
                "/dev/ttyUSB0",
                "flash",
                "--firmware",
                "tests/firmwares/skyconnect_zigbee_ncp_7.4.4.0.gbl",
            ],
            "/dev/ttyUSB0",
            DEFAULT_PROBE_METHODS,
            [],
        ),
        # With verbose flags (uses defaults)
        (
            [
                "-vvv",
                "--device",
                "/dev/ttyUSB1",
                "flash",
                "--firmware",
                "tests/firmwares/skyconnect_zigbee_ncp_7.4.4.0.gbl",
            ],
            "/dev/ttyUSB1",
            DEFAULT_PROBE_METHODS,
            [],
        ),
        # With single bootloader reset method
        (
            [
                "--device",
                "/dev/ttyUSB0",
                "--bootloader-reset",
                "rts_dtr",
                "flash",
                "--firmware",
                "tests/firmwares/skyconnect_zigbee_ncp_7.4.4.0.gbl",
            ],
            "/dev/ttyUSB0",
            DEFAULT_PROBE_METHODS,
            [ResetTarget.RTS_DTR],
        ),
        # With multiple bootloader reset methods (chained)
        (
            [
                "--device",
                "/dev/ttyUSB0",
                "--bootloader-reset",
                "rts_dtr,baudrate",
                "flash",
                "--firmware",
                "tests/firmwares/skyconnect_zigbee_ncp_7.4.4.0.gbl",
            ],
            "/dev/ttyUSB0",
            DEFAULT_PROBE_METHODS,
            [ResetTarget.RTS_DTR, ResetTarget.BAUDRATE],
        ),
        # With socket device
        (
            [
                "--device",
                "socket://192.168.1.100:1234",
                "flash",
                "--firmware",
                "tests/firmwares/skyconnect_zigbee_ncp_7.4.4.0.gbl",
            ],
            "socket://192.168.1.100:1234",
            DEFAULT_PROBE_METHODS,
            [],
        ),
    ],
)
async def test_flash_command_argument_parsing(
    args,
    expected_device,
    expected_probe_methods,
    expected_reset,
):
    """Test that flash command correctly parses various argument combinations."""
    result = await invoke_main(args)

    assert result.exit_code == 0
    assert isinstance(result.flasher, Flasher)

    assert result.flasher._device == expected_device
    assert set(result.flasher._probe_methods) == set(expected_probe_methods)
    assert result.flasher._reset_targets == expected_reset


@pytest.mark.parametrize(
    "args,expected_device",
    [
        (["--device", "/dev/ttyUSB0", "probe"], "/dev/ttyUSB0"),
        (["-v", "--device", "/dev/ttyUSB1", "probe"], "/dev/ttyUSB1"),
        (["--device", "socket://localhost:5000", "probe"], "socket://localhost:5000"),
    ],
)
async def test_probe_command_argument_parsing(args, expected_device):
    """Test that probe command correctly parses arguments."""
    result = await invoke_main(args)

    assert result.exit_code == 0
    assert result.flasher is not None
    assert result.flasher._device == expected_device


@pytest.mark.parametrize(
    "args,expected_ieee,expected_force",
    [
        (
            [
                "--device",
                "/dev/ttyUSB0",
                "write-ieee",
                "--ieee",
                "11:22:33:44:55:66:77:88",
            ],
            "11:22:33:44:55:66:77:88",
            False,
        ),
        (
            [
                "--device",
                "/dev/ttyUSB0",
                "write-ieee",
                "--ieee",
                "11:22:33:44:55:66:77:88",
                "--force",
            ],
            "11:22:33:44:55:66:77:88",
            True,
        ),
    ],
)
async def test_write_ieee_command_argument_parsing(args, expected_ieee, expected_force):
    """Test that write-ieee command correctly parses arguments."""
    with patch(
        "universal_silabs_flasher.flasher.Flasher.write_emberznet_eui64",
        new_callable=AsyncMock,
    ) as mock_write:
        result = await invoke_main(args)

        assert result.exit_code == 0

        mock_write.assert_called_once()
        call_args = mock_write.call_args

        assert str(call_args.args[0]) == expected_ieee
        assert call_args.kwargs["force"] == expected_force


async def test_dump_gbl_metadata_command():
    """Test that dump-gbl-metadata command works without --device."""
    result = await invoke_main(
        [
            "dump-gbl-metadata",
            "--firmware",
            "tests/firmwares/skyconnect_zigbee_ncp_7.4.4.0.gbl",
        ]
    )

    assert result.exit_code == 0
    assert '{"' in result.output or result.output.strip().endswith("null")


async def test_flash_profile_uses_registered_flasher():
    result = await invoke_main(
        [
            "--device",
            "/dev/ttyUSB0",
            "flash",
            "--profile",
            "zbt2",
            "--firmware",
            "tests/firmwares/skyconnect_zigbee_ncp_7.4.4.0.gbl",
        ]
    )

    assert result.exit_code == 0
    assert isinstance(result.flasher, Zbt2Flasher)


@pytest.mark.parametrize(
    "args,expected_error_fragment",
    [
        # Invalid bootloader reset method
        (
            [
                "--device",
                "/dev/ttyUSB0",
                "--bootloader-reset",
                "invalid_method",
                "flash",
                "--firmware",
                "tests/firmwares/skyconnect_zigbee_ncp_7.4.4.0.gbl",
            ],
            "invalid",
        ),
        # Invalid bootloader reset method in chain
        (
            [
                "--device",
                "/dev/ttyUSB0",
                "--bootloader-reset",
                "rts_dtr,invalid_method",
                "flash",
                "--firmware",
                "tests/firmwares/skyconnect_zigbee_ncp_7.4.4.0.gbl",
            ],
            "invalid",
        ),
        # Invalid probe method
        (
            [
                "--device",
                "/dev/ttyUSB0",
                "--probe-method",
                "invalid_app",
                "flash",
                "--firmware",
                "tests/firmwares/skyconnect_zigbee_ncp_7.4.4.0.gbl",
            ],
            "invalid",
        ),
        # Missing firmware for flash
        (
            ["--device", "/dev/ttyUSB0", "flash"],
            "required",
        ),
        # Missing IEEE for write-ieee
        (
            ["--device", "/dev/ttyUSB0", "write-ieee"],
            "required",
        ),
        # Removed --baudrate flag
        (
            [
                "--device",
                "/dev/ttyUSB0",
                "--baudrate",
                "115200",
                "flash",
                "--firmware",
                "tests/firmwares/skyconnect_zigbee_ncp_7.4.4.0.gbl",
            ],
            "error",
        ),
        (
            [
                "--device",
                "/dev/ttyUSB0",
                "--probe-methods",
                "ezsp:115200",
                "flash",
                "--profile",
                "zbt2",
                "--firmware",
                "tests/firmwares/skyconnect_zigbee_ncp_7.4.4.0.gbl",
            ],
            "--profile cannot be used",
        ),
        (
            [
                "--device",
                "/dev/ttyUSB0",
                "--bootloader-reset",
                "rts_dtr",
                "flash",
                "--profile",
                "zbt2",
                "--firmware",
                "tests/firmwares/skyconnect_zigbee_ncp_7.4.4.0.gbl",
            ],
            "--profile cannot be used",
        ),
    ],
)
async def test_invalid_argument_combinations_with_mocked_device(
    args, expected_error_fragment
):
    """Test invalid argument combinations with mocked device validator."""
    result = await invoke_main(args)

    assert result.exit_code != 0
    combined = result.output.lower() + result.stderr.lower()
    assert expected_error_fragment.lower() in combined


def make_zip(entries: dict[str, bytes]) -> bytes:
    buf = io.BytesIO()

    with zipfile.ZipFile(buf, "w") as zf:
        for name, data in entries.items():
            zf.writestr(name, data)

    return buf.getvalue()


async def test_firmware_download_gbl():
    """A plain .gbl downloaded over HTTP is flashed successfully."""
    with aioresponses() as m:
        m.get(FIRMWARE_URL, status=200, body=FIRMWARE_PATH.read_bytes())
        result = await invoke_main(
            ["--device", "/dev/ttyUSB0", "flash", "--firmware", FIRMWARE_URL]
        )

    assert result.exit_code == 0


async def test_firmware_download_zip_picks_gbl():
    """A ZIP with a .gbl and another file: the .gbl entry is extracted and flashed."""
    zip_url = "https://example.com/firmware/update.zip"
    zip_bytes = make_zip(
        {
            "readme.txt": b"hello",
            "firmware.gbl": FIRMWARE_PATH.read_bytes(),
        }
    )

    with aioresponses() as m:
        m.get(zip_url, status=200, body=zip_bytes)
        result = await invoke_main(
            ["--device", "/dev/ttyUSB0", "flash", "--firmware", zip_url]
        )

    assert result.exit_code == 0


async def test_firmware_download_failure():
    """An HTTP error status is reported as a CLI error."""
    with aioresponses() as m:
        m.get(FIRMWARE_URL, status=404)
        result = await invoke_main(
            ["--device", "/dev/ttyUSB0", "flash", "--firmware", FIRMWARE_URL]
        )

    assert result.exit_code != 0
    assert "error" in result.stderr.lower()



from __future__ import annotations

import asyncio
import dataclasses
import enum
import logging
import re
import typing

from zigpy.serial import SerialProtocol
import zigpy.types

from .common import PROBE_TIMEOUT, StateMachine, Version, asyncio_timeout, crc16_ccitt

_LOGGER = logging.getLogger(__name__)


class UploadError(Exception):
    pass


class NoFirmwareError(Exception):
    pass


class ReceiverCancelled(UploadError):
    """Receiver cancelled the transmission with a `CAN` status."""


MENU_AFTER_UPLOAD_TIMEOUT = 1.0
RUN_APPLICATION_DELAY = 2.0
XMODEM_BLOCK_SIZE = 128
XMODEM_RECEIVE_TIMEOUT = 2


class XModemPacketType(zigpy.types.enum8):
    """XModem packet type byte."""

    SOH = 0x01  # Start of Header
    EOT = 0x04  # End of Transmission
    CAN = 0x18  # Cancel
    ETB = 0x17  # End of Transmission Block
    ACK = 0x06  # Acknowledge
    NAK = 0x15  # Not Acknowledge


@dataclasses.dataclass(frozen=True)
class XmodemCRCPacket:
    """XModem CRC packet implementing the zigpy `serialize` API."""

    number: zigpy.types.uint8_t
    payload: bytes

    def serialize(self) -> bytes:
        """Serialize the packet, computing header and payload checksums."""
        assert len(self.payload) == XMODEM_BLOCK_SIZE
        return (
            bytes([XModemPacketType.SOH, self.number, 0xFF - self.number])
            + self.payload
            + crc16_ccitt(self.payload).to_bytes(2, "big")
        )


MENU_REGEX = re.compile(
    rb"\r\n(?P<type>Gecko|\w+ Serial) Bootloader v(?P<version>.*?)\r\n"
    rb"1\. upload (?:gbl|ebl)\r\n"
    rb"2\. run\r\n"
    rb"3\. ebl info\r\n"
    rb"(\d+\. .*?\r\n)*"  # All other options are ignored but we still expect a menu
    rb"BL > "
)

UPLOAD_STATUS_REGEX = re.compile(
    rb"\r\nSerial upload (?P<status>complete|aborted)\r\n"
    rb"(?P<message>.*?)\x00?",
    flags=re.DOTALL,
)  # fmt: skip


class State(str, enum.Enum):
    WAITING_FOR_MENU = "waiting_for_menu"
    IN_MENU = "in_menu"
    WAITING_XMODEM_READY = "waiting_xmodem_ready"
    XMODEM_UPLOADING = "xmodem_uploading"
    WAITING_UPLOAD_DONE = "waiting_upload_done"
    UPLOAD_DONE = "upload_done"


class GeckoBootloaderOption(bytes, enum.Enum):
    UPLOAD_FIRMWARE = b"1"
    RUN_FIRMWARE = b"2"
    EBL_INFO = b"3"


class GeckoBootloaderProtocol(SerialProtocol):
    _buffer: bytearray

    def __init__(self) -> None:
        super().__init__()
        self._state_machine = StateMachine(
            states={s.value for s in State},
            initial=State.WAITING_FOR_MENU,
        )
        self._version: str | None = None
        self._upload_status: str | None = None
        self.loop: asyncio.AbstractEventLoop | None = None

        # XMODEM state
        self._xmodem_firmware: bytes | None = None
        self._xmodem_chunk_index: int = 0
        self._xmodem_total_chunks: int = 0
        self._xmodem_retries: int = 0
        self._xmodem_max_retries: int = 0
        self._xmodem_progress_callback: (
            typing.Callable[[int, int], typing.Any] | None
        ) = None
        self._xmodem_completion_future: asyncio.Future[None] | None = None
        self._xmodem_timeout_handle: asyncio.TimerHandle | None = None

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        super().connection_made(transport)
        self.loop = asyncio.get_running_loop()

    def connection_lost(self, exc: BaseException | None) -> None:
        super().connection_lost(exc)
        self._state_machine.cancel_all_futures(
            exc or RuntimeError("Connection has been lost")
        )

        if self._xmodem_completion_future and not self._xmodem_completion_future.done():
            self._xmodem_completion_future.set_exception(
                exc or RuntimeError("Connection has been lost")
            )

        if self._xmodem_timeout_handle:
            self._xmodem_timeout_handle.cancel()

    async def probe(self) -> Version:
        """Attempt to communicate with the bootloader."""
        async with asyncio_timeout(PROBE_TIMEOUT):
            return await self.ebl_info()

    async def ebl_info(self) -> Version:
        """Select `ebl info` in the menu and return the bootloader version."""
        self._state_machine.state = State.WAITING_FOR_MENU

        # Ember bootloader requires a newline
        self.send_data(b"\n")
        self.send_data(GeckoBootloaderOption.EBL_INFO)

        await self._state_machine.wait_for_state(State.IN_MENU)

        assert self._version is not None
        return Version(self._version)

    async def run_firmware(self) -> None:
        """Select `run` in the menu."""
        await self._state_machine.wait_for_state(State.IN_MENU)

        # If the firmware fails to launch, the menu will appear again
        self._state_machine.state = State.WAITING_FOR_MENU
        self.send_data(GeckoBootloaderOption.RUN_FIRMWARE)

        try:
            async with asyncio_timeout(RUN_APPLICATION_DELAY):
                await self._state_machine.wait_for_state(State.IN_MENU)
        except asyncio.TimeoutError:
            # The menu did not appear so the application must be running
            return
        else:
            raise NoFirmwareError("No firmware exists on the device")

    def _xmodem_timeout_cb(self) -> None:
        """XMODEM receive timeout."""
        self._xmodem_timeout_handle = None
        _LOGGER.debug("XMODEM receive timeout")
        self._xmodem_retry_chunk()

    def _xmodem_retry_chunk(self) -> None:
        """Retry sending the current XMODEM chunk."""
        if self._xmodem_retries >= self._xmodem_max_retries:
            self._xmodem_abort(
                UploadError(f"Received {self._xmodem_max_retries} consecutive failures")
            )
            return

        self._xmodem_retries += 1
        _LOGGER.debug(
            "Retrying chunk %d (attempt %d)",
            self._xmodem_chunk_index,
            self._xmodem_retries,
        )
        self._xmodem_send_chunk_or_eot()

    def _xmodem_abort(self, exc: Exception) -> None:
        """Abort XMODEM transfer."""
        if self._xmodem_completion_future and not self._xmodem_completion_future.done():
            self._xmodem_completion_future.set_exception(exc)

        if self._xmodem_timeout_handle:
            self._xmodem_timeout_handle.cancel()
            self._xmodem_timeout_handle = None

    def _xmodem_send_chunk_or_eot(self) -> None:
        """Send the current XMODEM chunk or EOT if done."""
        if self._xmodem_chunk_index >= self._xmodem_total_chunks:
            _LOGGER.debug("Sending EOT")
            self.send_data(bytes([XModemPacketType.EOT]))
        else:
            _LOGGER.debug("Sending chunk %d", self._xmodem_chunk_index)
            assert self._xmodem_firmware is not None
            packet = XmodemCRCPacket(
                number=zigpy.types.uint8_t((self._xmodem_chunk_index + 1) & 0xFF),  # type: ignore[no-untyped-call]
                payload=self._xmodem_firmware[
                    XMODEM_BLOCK_SIZE * self._xmodem_chunk_index : XMODEM_BLOCK_SIZE
                    * (self._xmodem_chunk_index + 1)
                ],
            )
            self.send_data(packet.serialize())

        assert self.loop is not None
        self._xmodem_timeout_handle = self.loop.call_later(
            XMODEM_RECEIVE_TIMEOUT, self._xmodem_timeout_cb
        )

    async def upload_firmware(
        self,
        firmware: bytes,
        *,
        max_failures: int = 3,
        progress_callback: typing.Callable[[int, int], typing.Any] | None = None,
    ) -> None:
        """Select `upload gbl` in the menu and upload GBL firmware."""
        await self.ebl_info()

        # Select the option
        self._state_machine.state = State.WAITING_XMODEM_READY
        self.send_data(GeckoBootloaderOption.UPLOAD_FIRMWARE)

        # Wait for the XMODEM `C` byte
        await self._state_machine.wait_for_state(State.XMODEM_UPLOADING)

        # Set up XMODEM state
        self._xmodem_firmware = firmware
        self._xmodem_chunk_index = 0
        self._xmodem_total_chunks = len(firmware) // XMODEM_BLOCK_SIZE
        self._xmodem_retries = 0
        self._xmodem_max_retries = max_failures
        self._xmodem_progress_callback = progress_callback
        assert self.loop is not None
        self._xmodem_completion_future = self.loop.create_future()

        if self._xmodem_progress_callback is not None:
            self._xmodem_progress_callback(0, len(self._xmodem_firmware))

        # Start the transfer
        self._xmodem_send_chunk_or_eot()

        # Wait for transfer to complete
        await self._xmodem_completion_future

        # Clean up XMODEM state
        self._xmodem_firmware = None
        self._xmodem_completion_future = None

        # After XMODEM completes, data_received processes the upload status message
        # and transitions: UPLOAD_DONE -> WAITING_FOR_MENU -> IN_MENU.
        # (if menu is buffered). The menu is sometimes sent immediately after upload.
        try:
            async with asyncio_timeout(MENU_AFTER_UPLOAD_TIMEOUT):
                await self._state_machine.wait_for_state(State.IN_MENU)
        except asyncio.TimeoutError:
            # If not, trigger it manually
            await self.ebl_info()

        if self._upload_status != "complete":
            raise UploadError(self._upload_status)

    def send_data(self, data: bytes) -> None:
        assert self._transport is not None
        _LOGGER.debug("Sending data %s", data)
        self._transport.write(data)

    def _handle_xmodem_response(self) -> None:
        """Handle a single byte response from the receiver."""
        if not self._buffer:
            return

        # If we are not waiting for a response, this is spurious data
        if self._xmodem_timeout_handle is None:
            _LOGGER.debug("Ignoring spurious XMODEM data: %r", self._buffer)
            self._buffer.clear()
            return

        # We are waiting for a response, so cancel the timeout
        self._xmodem_timeout_handle.cancel()
        self._xmodem_timeout_handle = None

        response = self._buffer[0]
        self._buffer = self._buffer[1:]

        if response == XModemPacketType.ACK:
            if self._xmodem_chunk_index >= self._xmodem_total_chunks:
                # EOT was ACKed - keep buffer intact as upload complete/menu may follow
                if (
                    self._xmodem_completion_future
                    and not self._xmodem_completion_future.done()
                ):
                    self._xmodem_completion_future.set_result(None)
                self._state_machine.state = State.WAITING_UPLOAD_DONE
                # Process the rest of the buffer for the upload complete message
                self.data_received(b"")
                return

            assert self._xmodem_firmware is not None
            offset = (self._xmodem_chunk_index + 1) * XMODEM_BLOCK_SIZE
            if self._xmodem_progress_callback is not None:
                self._xmodem_progress_callback(offset, len(self._xmodem_firmware))

            _LOGGER.debug(
                "Firmware upload progress: %0.2f%%",
                100 * offset / len(self._xmodem_firmware),
            )

            self._xmodem_chunk_index += 1
            self._xmodem_retries = 0
            self._xmodem_send_chunk_or_eot()
        elif response == XModemPacketType.NAK:
            _LOGGER.debug("Got a NAK, retrying")
            self._xmodem_retry_chunk()
        elif response == XModemPacketType.CAN:
            self._xmodem_abort(ReceiverCancelled())
        else:
            _LOGGER.warning("Invalid XMODEM response: %r", response)
            # Treat as a failure and retry
            self._xmodem_retry_chunk()

    def data_received(self, data: bytes) -> None:
        super().data_received(data)

        if self._state_machine.state == State.XMODEM_UPLOADING:
            self._handle_xmodem_response()
            return

        while self._buffer:
            _LOGGER.debug("Parsing %s: %r", self._state_machine.state, self._buffer)
            current_state = self._state_machine.state

            if current_state == State.WAITING_FOR_MENU:
                match = MENU_REGEX.search(self._buffer)

                if match is None:
                    break

                self._version = match.group("version").decode("ascii")
                _LOGGER.debug("Detected version string %r", self._version)

                self._buffer.clear()
                self._state_machine.state = State.IN_MENU
            elif current_state == State.WAITING_XMODEM_READY:
                if not self._buffer.endswith(b"C"):
                    break

                self._buffer.clear()
                self._state_machine.state = State.XMODEM_UPLOADING
            elif current_state == State.WAITING_UPLOAD_DONE:
                match = UPLOAD_STATUS_REGEX.search(self._buffer)

                if match is None:
                    break

                status = match.group("status").decode("ascii")

                if status == "complete":
                    self._upload_status = status
                else:
                    self._upload_status = match.group("message").decode("ascii")

                del self._buffer[: match.span()[1]]
                self._state_machine.state = State.UPLOAD_DONE

                _LOGGER.debug("Upload status: %s", self._upload_status)
            elif current_state == State.UPLOAD_DONE:
                # Transition to waiting for menu and continue processing buffer
                self._state_machine.state = State.WAITING_FOR_MENU

            # If the state changed, re-evaluate the loop with the new state
            if self._state_machine.state != current_state:
                continue

            # Otherwise, we need more data
            break
